#!/usr/bin/env python
# encoding: utf-8

"""Azkaban CLI: a lightweight command line interface for Azkaban.

Usage:
  azkaban build [-cp PROJECT] [-a ALIAS | -u URL | [-r] ZIP] [-o OPTION ...]
  azkaban info [-p PROJECT] [-f | -o OPTION ... | [-i] JOB ...]
  azkaban log [-a ALIAS | -u URL] EXECUTION [JOB]
  azkaban run [-bkp PROJECT] [-a ALIAS | -u URL] [-e EMAIL ...]
              [-o OPTION ...] FLOW [JOB ...]
  azkaban schedule [-bkp PROJECT] [-a ALIAS | -u URL] [-e EMAIL ...]
                   [-o OPTION ...] [-s SPAN] (-d DATE) (-t TIME)
                   FLOW [JOB ...]
  azkaban upload [-cp PROJECT] [-a ALIAS | -u URL] ZIP
  azkaban -h | --help | -l | --log | -v | --version

Commmands:
  build*                        Build project and upload to Azkaban or save
                                locally the resulting archive.
  info*                         View information about jobs or files.
  log                           View workflow or job execution logs.
  run                           Run jobs or workflows. If no job is specified,
                                the entire workflow will be executed.
  schedule                      Schedule a workflow to be run at a specified
                                date and time.
  upload                        Upload archive to Azkaban server.

Arguments:
  EXECUTION                     Execution ID.
  JOB                           Job name.
  FLOW                          Workflow name. Recall that in the Azkaban world
                                this is simply a job without children.
  ZIP                           For `upload` command, the path to an existing
                                project zip archive. For `build`, the path
                                where the output archive will be built. If it
                                points to a directory, the archive will be
                                named after the project name (and version, if
                                present) and created in said directory.

Options:
  -a ALIAS --alias=ALIAS        Alias to saved URL and username. Will also try
                                to reuse session IDs for later connections.
  -b --bounce                   Skip execution if workflow is already running.
  -c --create                   Create the project if it does not exist.
  -d DATE --date=DATE           Date used for first run of a schedule. It must
                                be in the format `MM/DD/YYYY`.
  -e EMAIL --email=EMAIL        Email address to be notified when the workflow
                                finishes (can be specified multiple times).
  -f --files                    List project files instead of jobs. The first
                                column is the local path of the file, the
                                second the path of the file in the archive.
  -h --help                     Show this message and exit.
  -i --include-properties       Include project properties with job options.
  -k --kill                     Kill worfklow on first job failure.
  -l --log                      Show path to current log file and exit.
  -o OPTION --option=OPTION     Azkaban properties. Can either be the path to
                                a properties file or a single parameter
                                formatted as `key=value`, e.g. `-o
                                user.to.proxy=foo`. For the `build` and `run`
                                commands, these will be added to the project's
                                or run's properties respectively (potentially
                                overriding existing ones). For the `info`
                                command, this will cause only jobs with these
                                exact parameters to be displayed.
  -p PROJECT --project=PROJECT  Azkaban project. Can either be a project name
                                or a path to a python module/package defining
                                an `azkaban.Project` instance. Commands which
                                are followed by an asterisk will only work in
                                the latter case. If multiple projects are
                                registered, you can disambiguate as follows:
                                `--project=module:project_name`.
  -r --replace                  Overwrite any existing file.
  -s SPAN --span=SPAN           Period to repeat the scheduled flow. Must be
                                in format `1d`, a combination of magnitude and
                                unit of repetition. If not specified, the flow
                                will be run only once.
  -t TIME --time=TIME           Time when a schedule should be run. Must be of
                                the format `hh,mm,(AM|PM),(PDT|UTC|..)`.
  -u URL --url=URL              Azkaban endpoint (with protocol, and optionally
                                a username): '[user@]protocol:endpoint'. E.g.
                                'http://azkaban.server'. The username defaults
                                to the current user, as determined by `whoami`.
                                If you often use the same url, consider using
                                the `--alias` option instead.
  -v --version                  Show version and exit.

Azkaban CLI returns with exit code 1 if an error occurred and 0 otherwise.

"""

from azkaban import __version__, CLI_ARGS
from azkaban.project import Project
from azkaban.remote import Execution, Session
from azkaban.util import (AzkabanError, Config, catch, flatten, human_readable,
  temppath, read_properties, write_properties)
from docopt import docopt
from requests.exceptions import HTTPError
import logging as lg
import os
import os.path as osp
import sys


_logger = lg.getLogger(__name__)


def _forward(args, names):
  """Forward subset of arguments from initial dictionary.

  :param args: Dictionary of parsed arguments (output of `docopt.docopt`).
  :param names: List of names that will be included.

  """
  names = set(names)
  return dict(
    ('_%s' % (k.lower().lstrip('-').replace('-', '_'), ), v)
    for (k, v) in args.items() if k in names
  )

def _parse_option(_option):
  """Parse `--option` argument.

  :param _option: `--option` argument.

  Returns a dictionary.

  """
  paths = (opt for opt in _option if not '=' in opt)
  opts = read_properties(*paths)
  try:
    opts.update(dict(s.split('=', 1) for s in _option if '=' in s))
  except ValueError:
    raise AzkabanError('Invalid `--option` flag.')
  return opts

def _parse_project(_project, require_project=False):
  """Parse `--project` argument into `(name, project)`.

  :param _project: `--project` argument.
  :param require_project: Fail if we fail to load the project.

  Note that `name` is guaranteed to be non-`None` (this function will throw an
  exception otherwise) but `project` can be.

  The rules are as follows:

  + If at least one `':'` is found in `_project` then the rightmost one is
    interpreted as delimitor between the path to the module and the project
    name.

  + Else:

    + We first try to interpret `_project` as a module path and find a unique
      project inside.

    + If the above attempt raises an `ImportError`, we interpret it as a name.

  """
  default_module = Config().get_option('azkaban', 'project', 'jobs')
  projects = {}
  _project = _project or default_module
  if ':' in _project:
    # unambiguous case
    path, name = _project.rsplit(':', 1)
    try:
      projects = Project.load(path or default_module)
      # adding the default here lets options like `-p :name` work as intended
    except ImportError:
      pass
  else:
    # the option could be a name or module
    try:
      # try first as a module
      projects = Project.load(_project)
    except ImportError:
      # if that fails, try as a name: load the default module and look there
      name = _project
      try:
        projects = Project.load(default_module)
      except ImportError:
        pass
    else:
      name = None
  if name:
    if name in projects:
      return name, projects[name]
    elif projects:
      # harder consistency requirement
      raise AzkabanError(
        'Project %r not found. Available projects: %s\n'
        'You can also specify another location using the `--project` option.'
        % (name, ', '.join(projects))
      )
    elif require_project:
      raise AzkabanError(
        'This command requires a project configuration module.\n'
        'You can specify another location using the `--project` option.'
      )
    else:
      return name, None
  else:
    if not projects:
      raise AzkabanError(
        'No registered project found in %r.\n'
        'You can also specify another location using the `--project` option.'
        % (_project, )
      )
    elif len(projects) > 1:
      raise AzkabanError(
        'Multiple registered projects found: %s\n'
        'You can use the `--project` option to disambiguate.'
        % (', '.join(projects), )
      )
    else:
      return projects.popitem()

def _get_project_name(_project):
  """Return project name.

  :param _project: `--project` argument.

  """
  return _parse_project(_project)[0]

def _load_project(_project):
  """Resolve project from CLI argument.

  :param _project: `--project` argument.

  """
  try:
    name, project = _parse_project(_project, require_project=True)
  except ImportError:
    raise AzkabanError(
      'This command requires a project configuration module which was not '
      'found.\nYou can specify another location using the `--project` option.'
    )
  else:
    return project

def _upload_callback(cur_bytes, tot_bytes, file_index, _stdout=sys.stdout):
  """Callback for streaming upload.

  :param cur_bytes: Total bytes uploaded so far.
  :param tot_bytes: Total bytes to be uploaded.
  :param file_index: (0-based) index of the file currently uploaded.
  :param _stdout: Performance caching.

  """
  if cur_bytes != tot_bytes:
    _stdout.write(
      'Uploading project: %.1f%%\r'
      % (100. * cur_bytes / tot_bytes, )
    )
  else:
    _stdout.write('Validating project...    \r')
  _stdout.flush()

def view_info(project, _files, _option, _job, _include_properties):
  """List jobs in project."""
  if _job:
    if _include_properties:
      write_properties(
        flatten(project.properties),
        header='project.properties'
      )
    for name in _job:
      project.jobs[name].build(header='%s.job' % (name, ))
  elif _files:
    for path, archive_path in sorted(project.files):
      sys.stdout.write('%s\t%s\n' % (osp.relpath(path), archive_path))
  else:
    options = _parse_option(_option).items()
    jobs = sorted(project.jobs.items())
    dependencies = set(
      dep
      for _, job in jobs
      for dep in job.options.get('dependencies', '').split(',')
    )
    for name, job in jobs:
      if all(job.options.get(k) == v for k, v in options):
        sys.stdout.write(
          '%s\t%s\n'
          % ('J' if name in dependencies else 'F', name, )
        )

def view_log(_execution, _job, _url, _alias):
  """View workflow or job execution logs."""
  session = Session(_url, _alias)
  exc = Execution(session, _execution)
  logs = exc.job_logs(_job[0]) if _job else exc.logs()
  try:
    for line in logs:
      sys.stdout.write('%s\n' % (line.encode('utf-8'), ))
  except HTTPError:
    # Azkaban responds with 500 if the execution or job isn't found
    if _job:
      raise AzkabanError(
        'Execution %s and/or job %s not found.', _execution, _job
      )
    else:
      raise AzkabanError('Execution %s not found.', _execution)

def run_workflow(project_name, _flow, _job, _url, _alias, _bounce, _kill,
  _email, _option):
  """Run workflow."""
  session = Session(_url, _alias)
  res = session.run_workflow(
    name=project_name,
    flow=_flow,
    jobs=_job,
    concurrent=not _bounce,
    on_failure='cancel' if _kill else 'finish',
    emails=_email,
    properties=_parse_option(_option),
  )
  exec_id = res['execid']
  job_names = ', jobs: %s' % (', '.join(_job), ) if _job else ''
  sys.stdout.write(
    'Flow %s successfully submitted (execution id: %s%s).\n'
    'Details at %s/executor?execid=%s\n'
    % (_flow, exec_id, job_names, session.url, exec_id)
  )

def schedule_workflow(project_name, _date, _time, _span, _flow, _job, _url,
  _alias, _bounce, _kill, _email, _option):
  """Schedule workflow."""
  session = Session(_url, _alias)
  session.schedule_workflow(
    name=project_name,
    flow=_flow,
    date=_date,
    time=_time,
    period=_span,
    jobs=_job,
    concurrent=not _bounce,
    on_failure='cancel' if _kill else 'finish',
    emails=_email,
    properties=_parse_option(_option),
  )
  sys.stdout.write(
    'Flow %s scheduled successfully.\n' % (_flow, )
  )

def upload_project(project_name, _zip, _url, _alias, _create):
  """Upload project."""
  session = Session(_url, _alias)
  while True:
    try:
      res = session.upload_project(
        name=project_name,
        path=_zip,
        callback=_upload_callback
      )
    except AzkabanError as err:
      if _create:
        session.create_project(project_name, project_name)
      else:
        raise err
    else:
      break
  sys.stdout.write(
    'Project %s successfully uploaded (id: %s, size: %s, version: %s).\n'
    'Details at %s/manager?project=%s\n'
    % (
      project_name,
      res['projectId'],
      human_readable(osp.getsize(_zip)),
      res['version'],
      session.url,
      project_name,
    )
  )

def build_project(project, _zip, _url, _alias, _replace, _create, _option):
  """Build project."""
  if _option:
    project.properties = flatten(project.properties)
    # to make sure we properly override nested options, we flatten first
    project.properties.update(_parse_option(_option))
  if _zip:
    if osp.isdir(_zip):
      _zip = osp.join(_zip, '%s.zip' % (project.versioned_name, ))
    project.build(_zip, overwrite=_replace)
    sys.stdout.write(
      'Project %s successfully built and saved as %r (size: %s).\n'
      % (project, _zip, human_readable(osp.getsize(_zip)))
    )
  else:
    with temppath() as _zip:
      project.build(_zip)
      archive_name = '%s.zip' % (project.versioned_name, )
      session = Session(_url, _alias)
      while True:
        try:
          res = session.upload_project(
            name=project.name,
            path=_zip,
            archive_name=archive_name,
            callback=_upload_callback
          )
        except AzkabanError as err:
          if _create and str(err).endswith("doesn't exist."):
            session.create_project(project.name, project.name)
          else:
            raise err
        else:
          break
      sys.stdout.write(
        'Project %s successfully built and uploaded '
        '(id: %s, size: %s, upload: %s).\n'
        'Details at %s/manager?project=%s\n'
        % (
          project,
          res['projectId'],
          human_readable(osp.getsize(_zip)),
          res['version'],
          session.url,
          project,
        )
      )

@catch(AzkabanError)
def main(argv=None):
  """Entry point."""
  # enable general logging
  logger = lg.getLogger()
  logger.setLevel(lg.DEBUG)
  handler = Config().get_file_handler('azkaban')
  if handler:
    logger.addHandler(handler)
  # capture pesky unverified requests warnings
  lg.captureWarnings(True)
  # parse arguments
  argv = argv or sys.argv[1:]
  _logger.debug('Running command %r from %r.', ' '.join(argv), os.getcwd())
  args = docopt(__doc__, version=__version__)
  CLI_ARGS.update(args)
  # do things
  if args['--log']:
    if handler:
      sys.stdout.write('%s\n' % (handler.baseFilename, ))
    else:
      raise AzkabanError('No log file active.')
  elif args['build']:
    build_project(
      _load_project(args['--project']),
      **_forward(
        args,
        ['ZIP', '--url', '--alias', '--replace', '--create', '--option']
      )
    )
  elif args['log']:
    view_log(
      **_forward(args, ['EXECUTION', 'JOB', '--url', '--alias'])
    )
  elif args['info']:
    view_info(
      _load_project(args['--project']),
      **_forward(args, ['--files', '--option', 'JOB', '--include-properties'])
    )
  elif args['run']:
    run_workflow(
      _get_project_name(args['--project']),
      **_forward(
        args,
        [
          'FLOW', 'JOB', '--bounce', '--url', '--alias', '--kill', '--email',
          '--option',
        ]
      )
    )
  elif args['schedule']:
    schedule_workflow(
      _get_project_name(args['--project']),
      **_forward(
        args,
        [
          'FLOW', 'JOB', '--bounce', '--url', '--alias', '--kill',
          '--email', '--option', '--date', '--time', '--span'
        ]
      )
    )
  elif args['upload']:
    upload_project(
      _get_project_name(args['--project']),
      **_forward(args, ['ZIP', '--create', '--url', '--alias'])
    )

if __name__ == '__main__':
  main()



#!/usr/bin/env python
# encoding: utf-8

"""Test Azkaban remote module."""

from azkaban.ext.pig import PigJob
from azkaban.project import Project
from azkaban.job import Job
from azkaban.remote import Execution, Session, _parse_url
from azkaban.util import AzkabanError, Config, temppath
from six.moves.configparser import NoOptionError, NoSectionError
from nose.tools import eq_, ok_, raises, nottest
from nose.plugins.skip import SkipTest
from time import sleep


class _TestSession(object):

  """Base class to run tests on an Azkaban server.

  These will be skipped if no valid credentials (url and associated session id)
  are found.

  If the class variable `project` is specified, the corresponding project
  will be created before each test.

  """

  project_name = None
  session = None

  @classmethod
  def setup_class(cls):
    if not cls.session:
      config = Config()
      try:
        alias = config.parser.get('azkaban', 'test.alias')
      except (NoOptionError, NoSectionError):
        pass
      else:
        cls.session = Session(alias=alias)
        try:
          cls.session.create_project(cls.project_name, 'Testing project.')
        except AzkabanError:
          pass # project already exists somehow

  @classmethod
  def teardown_class(cls):
    if cls.session:
      try:
        cls.session.delete_project(cls.project_name)
      except AzkabanError:
        pass # project was already deleted

  def setup(self):
    if not self.session:
      raise SkipTest
    if self.project_name:
      sleep(2)
      self.project = Project(self.project_name)


class TestCreateDelete(_TestSession):

  project_name = 'azkabancli_test_create_delete'

  @classmethod
  def setup_class(cls):
    # don't create project automatically (goal of these tests...)
    if not cls.session:
      config = Config()
      try:
        alias = config.parser.get('azkaban', 'test.alias')
      except (NoOptionError, NoSectionError):
        pass
      else:
        cls.session = Session(alias=alias)

  def project_exists(self, project):
    try:
      try:
        project.add_job('test', Job({'type': 'noop'}))
      except AzkabanError:
        pass # job was already added
      with temppath() as path:
        project.build(path)
        self.session.upload_project(project, path)
    except AzkabanError:
      return False
    else:
      return True

  def test_create_delete_project(self):
    ok_(not self.project_exists(self.project))
    self.session.create_project(self.project, 'Some description.')
    ok_(self.project_exists(self.project))
    self.session.delete_project(self.project)
    ok_(not self.project_exists(self.project))

  @raises(AzkabanError)
  def test_create_duplicate_project(self):
    self.session.create_project(self.project, 'Some description.')
    self.session.create_project(self.project, 'Some other description.')

  @raises(AzkabanError)
  def test_delete_nonexistent_project(self):
    if self.project_exists(self.project):
      self.session.delete_project(self.project)
    self.session.delete_project(self.project)


class TestUpload(_TestSession):

  project_name = 'azkabancli_test_upload'

  @raises(AzkabanError)
  def test_missing_archive(self):
    self.session.upload_project(self.project, 'foo')

  @raises(AzkabanError)
  def test_invalid_project(self):
    project = Project('an_non_existent_project')
    with temppath() as path:
      project.add_job('test', Job({'type': 'noop'}))
      project.build(path)
      self.session.upload_project(project, path)

  def test_upload_simple(self):
    with temppath() as path:
      self.project.add_job('test', Job({'type': 'noop'}))
      self.project.build(path)
      res = self.session.upload_project(self.project, path)
      eq_(['projectId', 'version'], sorted(res))

  @raises(AzkabanError)
  def test_upload_missing_type(self):
    with temppath() as path:
      self.project.add_job('test', Job())
      self.project.build(path)
      self.session.upload_project(self.project, path)

  def test_upload_pig_job(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('-- pig script')
      self.project.add_job('foo', PigJob({'pig.script': path}))
      with temppath() as path:
        self.project.build(path)
        res = self.session.upload_project(self.project, path)
    eq_(['projectId', 'version'], sorted(res))


class TestGetWorkflowInfo(_TestSession):

  project_name = 'azkaban_cli_test_flow_jobs'

  def get_job_names(self, flow_info):
    return [n['id'] for n in flow_info['nodes']]

  @raises(AzkabanError)
  def test_missing_project(self):
    self.session.get_workflow_info('some_missing_project', 'baz')

  @raises(AzkabanError)
  def test_invalid_flow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    self.session.get_workflow_info(self.project, 'baz')

  def test_get_single_job(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    info = self.session.get_workflow_info(self.project, 'foo')
    eq_(self.get_job_names(info), ['foo'])

  def test_get_multiple_jobs(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    info = self.session.get_workflow_info(self.project, 'bar')
    eq_(sorted(self.get_job_names(info)), ['bar', 'foo'])


class TestGetProjectInfo(_TestSession):

  project_name = 'azkaban_cli_test_project_info'

  def test_valid_name(self):
    project_id = self.session._get_project_id(self.project_name)
    project_id2 = self.session._get_project_id(self.project_name)
    ok_(isinstance(project_id, int))
    eq_(project_id, project_id2)

  @raises(AzkabanError)
  def test_missing_name(self):
    self.session._get_project_id('DoesNotExist')


class TestRun(_TestSession):

  project_name = 'azkaban_cli_run'

  def test_run_simple_workflow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    res = self.session.run_workflow(self.project, 'foo')
    eq_(['execid', 'flow', 'message', 'project'], sorted(res.keys()))
    eq_(res['message'][:32], 'Execution submitted successfully')

  def test_run_workflow_with_dependencies(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    res = self.session.run_workflow(self.project, 'bar')
    eq_(['execid', 'flow', 'message', 'project'], sorted(res.keys()))
    eq_(res['message'][:32], 'Execution submitted successfully')

  @raises(AzkabanError)
  def test_run_missing_workflow(self):
    self.session.run_workflow(self.project, 'foo2')

  @raises(AzkabanError)
  def test_run_non_workflow_job(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    res = self.session.run_workflow(self.project, 'foo')

  @raises(AzkabanError)
  def test_run_blocking_workflow(self):
    options = {'type': 'command', 'command': 'sleep 2'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    self.session.run_workflow(self.project, 'foo')
    self.session.run_workflow(self.project, 'foo', concurrent=False)

  def test_run_non_blocking_workflow(self):
    options = {'type': 'command', 'command': 'sleep 2'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    self.session.run_workflow(self.project, 'foo')
    res = self.session.run_workflow(self.project, 'foo')
    eq_(['execid', 'flow', 'message', 'project'], sorted(res.keys()))
    eq_(res['message'][:32], 'Flow foo is already running with')

  def test_run_fail_early(self):
    options = {'type': 'command', 'command': 'sleep ${time}'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'time': 5}))
    self.project.add_job('f', Job({'type': 'noop', 'dependencies': 'foo,bar'}))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    res = self.session.run_workflow(self.project, 'f', on_failure='cancel')
    exec_id = res['execid']
    sleep(2)
    eq_(self.session.get_execution_status(exec_id)['status'], 'FAILED')

  def test_run_fail_finish(self):
    options = {'type': 'command', 'command': 'sleep ${time}'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'time': 5}))
    self.project.add_job('f', Job({'type': 'noop', 'dependencies': 'foo,bar'}))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    res = self.session.run_workflow(self.project, 'f', on_failure='finish')
    eid = res['execid']
    sleep(2)
    eq_(self.session.get_execution_status(eid)['status'], 'FAILED_FINISHING')

  @raises(ValueError)
  def test_run_fail_invalid_on_failure(self):
    options = {'type': 'command', 'command': 'sleep ${time}'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'time': 5}))
    self.project.add_job('f', Job({'type': 'noop', 'dependencies': 'foo,bar'}))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    self.session.run_workflow(self.project, 'f', on_failure='foobar')

  @raises(AzkabanError)
  def test_run_wrong_job_in_workflow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    self.session.run_workflow(self.project, 'foo', jobs=['bar'])

  def test_run_single_job_in_workflow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    self.session.run_workflow(self.project, 'bar', jobs=['foo'])


class TestExecution(_TestSession):

  project_name = 'azkabancli_test_execution'

  def setup(self):
    super(TestExecution, self).setup()
    options = {'type': 'command', 'command': 'sleep 4'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)

  def test_execution_start(self):
    exe = Execution.start(self.session, self.project, 'foo')
    sleep(2)
    eq_(exe.status['status'], 'RUNNING')
    sleep(4)
    eq_(exe.status['status'],'SUCCEEDED')

  def test_execution_cancel(self):
    exe = Execution.start(self.session, self.project, 'foo')
    sleep(1)
    exe.cancel()
    sleep(1)
    eq_(exe.status['status'],'KILLED')

  def test_execution_logs(self):
    exe = Execution.start(self.session, self.project, 'foo')
    logs = '\n'.join(exe.logs(2))
    ok_('Submitting job \'foo\' to run.' in logs)


class TestGetWorkflowExecutions(_TestSession):

  project_name = 'azkabancli_test_get_workflow_executions'

  def setup(self):
    super(TestGetWorkflowExecutions, self).setup()
    options = {'type': 'command', 'command': 'sleep 4'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)

  @raises(AzkabanError)
  def test_missing_project(self):
    self.session.get_workflow_executions('missing_project', 'foobar')

  def test_missing_flow(self):
    d = self.session.get_workflow_executions(self.project, 'foobar')
    eq_(len(d['executions']), 0)
    eq_(d['project'], self.project_name)
    eq_(d['flow'], 'foobar')

  def test_show_executions(self):
    exe = Execution.start(self.session, self.project, 'foo')
    sleep(5)
    d = self.session.get_workflow_executions(self.project, 'foo')
    ok_(len(d['executions']))
    executions = [e for e in d['executions'] if e['execId'] == exe.exec_id]
    eq_(len(executions), 1)
    eq_(executions[0]['status'], 'SUCCEEDED')

  def test_show_running_executions(self):
    exe = Execution.start(self.session, self.project, 'foo')
    sleep(2)
    d = self.session.get_workflow_executions(self.project, 'foo')
    running = [e for e in d['executions'] if e['status'] == 'RUNNING']
    eq_(len(running), 1)
    eq_(running[0]['execId'], exe.exec_id)


class TestGetRunningWorkflows(_TestSession):

  project_name = 'azkabancli_test_get_running_workflows'

  def setup(self):
    super(TestGetRunningWorkflows, self).setup()
    options = {'type': 'command', 'command': 'sleep 4'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)

  @raises(AzkabanError)
  def test_missing_project(self):
    self.session.get_running_workflows('missing_project', 'foobar')

  def test_missing_flow(self):
    d = self.session.get_running_workflows(self.project, 'foobar')
    eq_(d, {})

  def test_no_running(self):
    d = self.session.get_running_workflows(self.project, 'foo')
    eq_(d, {})

  def test_running(self):
    exe = Execution.start(self.session, self.project, 'foo')
    sleep(2)
    d = self.session.get_running_workflows(self.project, 'foo')
    eq_(d, {'execIds': [exe.exec_id]})


class TestSchedule(_TestSession):

  project_name = 'azkabancli_test_schedule'

  def setup(self):
    super(TestSchedule, self).setup()
    options = {'type': 'command', 'command': 'sleep 4'}
    self.project.add_job('foo', Job(options))
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)

  def test_schedule_unschedule(self):
    res = self.session.schedule_workflow(
      self.project_name,
      'foo',
      '08/07/2014',
      '9,21,PM,PDT',
      '1d',
    )
    ok_(res['status'] == 'success')
    res = self.session.unschedule_workflow(self.project_name, 'foo')
    ok_(res['status'] == 'success')

  @raises(AzkabanError)
  def test_invalid_unschedule(self):
    res = self.session.unschedule_workflow('DoesNotExist', 'foo')

  @raises(AzkabanError)
  def test_invalid_schedule(self):
    res = self.session.schedule_workflow(
      self.project_name,
      'foo',
      '08/07/2014',
      '9,21PM,PDT',
      '1d',
    )


class TestProperties(_TestSession):

  project_name = 'azkabancli_test_properties'

  message = 'This is definitely a unique message.'
  override = 'This is even more definitely a unique message.'

  def _run_workflow(self, flow, **kwargs):
    with temppath() as path:
      self.project.build(path)
      self.session.upload_project(self.project, path)
    exe = Execution.start(self.session, self.project, flow, properties=kwargs)
    for i in range(5):
      # wait until workflow is launched
      sleep(1)
      try:
        status = exe.status
      except AzkabanError:
        pass
      else:
        if status['status'] != 'PREPARING':
          break
    return exe

  def _add_command_job(self, name, command, **kwargs):
    self.project.add_job(
      name,
      Job({'type': 'command', 'command': command}, kwargs),
    )

  def _add_flow_job(self, name, flow, **kwargs):
    self.project.add_job(
      name,
      Job({'type': 'flow', 'flow.name': flow}, kwargs),
    )

  def test_global_properties(self):
    self.project.properties = {'msg': self.message}
    self._add_command_job('foo', 'echo ${msg}')
    exe = self._run_workflow('foo')
    ok_(self.message in '\n'.join(exe.job_logs('foo', 1)))

  def test_missing_global_properties(self):
    self._add_command_job('foo', 'echo ${msg}')
    exe = self._run_workflow('foo')
    eq_(exe.status['status'], 'FAILED')

  def test_options_override_global_properties(self):
    self.project.properties = {'msg': self.message}
    self._add_command_job('foo', 'echo ${msg}', msg=self.override)
    exe = self._run_workflow('foo')
    ok_(self.override in '\n'.join(exe.job_logs('foo', 1)))

  def test_runtime_properties_override_global_properties(self):
    # runtime properties can be used to override .properties options
    self.project.properties = {'msg': self.message}
    self._add_command_job('foo', 'echo ${msg}')
    exe = self._run_workflow('foo', msg=self.override)
    ok_(self.override in '\n'.join(exe.job_logs('foo', 1)))

  def test_options_override_runtime_properties(self):
    # but runtime properties don't override .job options
    self._add_command_job('foo', 'echo ${msg}', msg=self.message)
    exe = self._run_workflow('foo', msg=self.override)
    ok_(self.message in '\n'.join(exe.job_logs('foo', 1)))

  def test_embedded_properties(self):
    # note the colon separated notation for embedded flows
    self._add_command_job('foo', 'echo ${msg}')
    self._add_flow_job('bar', 'foo', msg=self.message)
    exe = self._run_workflow('bar')
    ok_(self.message in '\n'.join(exe.job_logs('bar:foo', 1)))

  def test_embedded_properties_propagate(self):
    # embedded flow properties also propagate to nested flows
    self._add_command_job('foo', 'echo ${msg}')
    self._add_flow_job('flow1', 'foo')
    self._add_flow_job('flow2', 'flow1', msg=self.message)
    exe = self._run_workflow('flow2')
    ok_(self.message in '\n'.join(exe.job_logs('flow2:flow1:foo', 1)))

  def test_embedded_properties_override_propagated_embedded_properties(self):
    # embedded flow properties also propagate to nested flows
    self._add_command_job('foo', 'echo ${msg}')
    self._add_flow_job('flow1', 'foo', msg=self.override)
    self._add_flow_job('flow2', 'flow1', msg=self.message)
    exe = self._run_workflow('flow2')
    ok_(self.override in '\n'.join(exe.job_logs('flow2:flow1:foo', 1)))

  def test_embedded_properties_override_global_properties(self):
    self.project.properties = {'msg': self.message}
    self._add_command_job('foo', 'echo ${msg}')
    self._add_flow_job('bar', 'foo', msg=self.override)
    exe = self._run_workflow('bar')
    ok_(self.override in '\n'.join(exe.job_logs('bar:foo', 1)))

  def test_embedded_properties_override_runtime_properties(self):
    # embedded flow properties override runtime properties (!)
    self._add_command_job('foo', 'echo ${msg}')
    self._add_flow_job('bar', 'foo', msg=self.override)
    exe = self._run_workflow('bar', msg=self.message)
    ok_(self.override in '\n'.join(exe.job_logs('bar:foo', 1)))

  def test_options_override_embedded_properties(self):
    # embedded flow properties don't override job options
    self._add_command_job('foo', 'echo ${msg}', msg=self.override)
    self._add_flow_job('bar', 'foo', msg=self.message)
    exe = self._run_workflow('bar')
    ok_(self.override in '\n'.join(exe.job_logs('bar:foo', 1)))


class TestParseUrl(object):

  project_name = 'azkabancli_test_parse_url'

  def test_good_url(self):
    eq_(_parse_url('http://admin:abc123@server:8081'),
        ('admin', 'abc123', 'http://server:8081'))
    eq_(_parse_url('http://admin@server:8081'),
        ('admin', None, 'http://server:8081'))
    eq_(_parse_url('http://server:8081'),
        (None, None, 'http://server:8081'))

  def test_cred_first_url(self):
    eq_(_parse_url('admin:abc123@http://server:8081'),
        ('admin', 'abc123', 'http://server:8081'))
    eq_(_parse_url('admin@http://server:8081'),
        ('admin', None, 'http://server:8081'))

  def test_no_scheme_url(self):
    eq_(_parse_url('admin:abc123@server:8081'),
        ('admin', 'abc123', 'http://server:8081'))

  @raises(AzkabanError)
  def test_bad_cred_url(self):
    _parse_url('admin:abc123@http://bad:server:name:8081')

  @raises(Exception)
  def test_bad_cred_url(self):
    _parse_url('http://admin:abc123@bad:server:name:8081')



#!/usr/bin/env python
# encoding: utf-8

"""Job definition module."""

from .util import flatten, write_properties


class Job(object):

  """Base Azkaban job.

  :param options: tuple of dictionaries. The final job options are built from
    this tuple by keeping the latest definition of each option. Furthermore, by
    default, any nested dictionary will be flattened (combining keys with
    `'.'`). Both these features can be changed by simply overriding the job
    constructor.

  To enable more functionality, subclass and override the :meth:`on_add` and
  :meth:`build` methods. The :meth:`join_option` and :meth:`join_prefix`
  methods are also provided as helpers to write custom jobs.

  """

  def __init__(self, *options):
    self.options = {}
    for option in options:
      self.options.update(flatten(option))

  def build(self, path=None, header=None):
    """Write job file.

    :param path: Path where job file will be created. Any existing file will
      be overwritten. Writes to stdout if no path is specified.
    :param header: Optional comment to be included at the top of the job file.

    """
    write_properties(self.options, path=path, header=header)

  def on_add(self, project, name, **kwargs):
    """Handler called when the job is added to a project.

    :param project: :class:`~azkaban.project.Project` instance
    :param name: name corresponding to this job in the project.
    :param kwargs: Keyword arguments. If this method is triggered by
      :meth:`~azkaban.project.Project.add_job`, the latter's keyword arguments
      will simply be forwarded. Else if this method is triggered by a merge,
      kwargs will be a dictionary with single key `'merging'` and value the
      merged project.

    The default implementation does nothing.

    """
    pass

  def join_option(self, option, sep, formatter='%s'):
    """Helper method to join iterable options into a string.

    :param key: Option key. If the option doesn't exist, this method does
      nothing.
    :param sep: Separator used to concatenate the string.
    :param formatter: Pattern used to format the option values.

    Example usage:

    .. code-block:: python

      class MyJob(Job):

        def __init__(self, *options):
          super(MyJob, self).__init__(*options)
          self.join_option('dependencies', ',')

      # we can now use lists to define job dependencies
      job = MyJob({'type': 'noop', 'dependencies': ['bar', 'foo']})

    """
    values = self.options.get(option, None)
    if values:
      self.options[option] = sep.join(formatter % (v, ) for v in values)

  def join_prefix(self, prefix, sep, formatter):
    """Helper method to join options starting with a prefix into a string.

    :param prefix: Option prefix.
    :param sep: Separator used to concatenate the string.
    :param formatter: String formatter. It is formatted using the tuple
      `(suffix, value)` where `suffix` is the part of `key` after `prefix`.

    Example usage:

    .. code-block:: python

      class MyJob(Job):

        def __init__(self, *options):
          super(MyJob, self).__init__(*options)
          self.join_prefix('jvm.args', ' ', '-D%s=%s')

      # we can now define JVM args using nested dictionaries
      job = MyJob({'type': 'java', 'jvm.args': {'foo': 48, 'bar': 23}})

    """
    prefix = prefix.rstrip('.')
    opts = []
    for key in list(self.options):
      # copying keys to a list to allow modifying dict in loop
      # note that we aren't using `.keys()` for compatibility with python3
      if key.startswith(prefix):
        opts.append(
          (key.replace('%s.' % (prefix, ), ''), self.options.pop(key))
        )
    if opts:
      self.options[prefix] = sep.join(formatter % a for a in sorted(opts))



#!/usr/bin/env python
# encoding: utf-8

"""Project definition module."""

from weakref import WeakValueDictionary
from zipfile import ZipFile
from .util import AzkabanError, Adapter, flatten, temppath, write_properties
import logging as lg
import os
import os.path as osp
import sys


_logger = lg.getLogger(__name__)


class _JobDict(dict):

  """Simple dictionary subclass for jobs.

  It disables the default `__setitem__` method and implements a custom
  `KeyError` exception. Note that this isn't completely fool-proof but enough
  for our purpose.

  """

  def __getitem__(self, key):
    try:
      return super(_JobDict, self).__getitem__(key)
    except KeyError:
      raise AzkabanError('Job %r not found.', key)

  def __setitem__(self, key, value):
    raise AzkabanError('Cannot insert job. Use `Project.add_job` instead.')


class Project(object):

  """Azkaban project.

  :param name: Name of the project.
  :param register: Add project to registry. Setting this to `False` will make
    it invisible to the CLI.
  :param root: Path to a root file or directory used to enable adding files
    using relative paths (typically used with `root=__file__`).
  :param version: Project version, currently only used for setting the name
    of the archive uploaded to Azkaban.

  The `properties` attribute of a project is a dictionary which can be used to
  pass Azkaban options which will then be available to all jobs in the project.
  This can be used for example to set project wide defaults.

  To avoid undefined behavior, both the `name` and `root` attributes should not
  be altered after instantiation.

  """

  root = None
  _registry = WeakValueDictionary()

  def __init__(self, name, root=None, register=True, version=None):
    self.name = name
    self.version = version
    if root:
      self.root = osp.abspath(root if osp.isdir(root) else osp.dirname(root))
    if register:
      self._registry[name] = self
    self._jobs = {}
    self._files = {}
    self.properties = {}
    self._logger = Adapter(repr(self), _logger)
    self._logger.debug('Instantiated.')

  def __repr__(self):
    return '<%s(name=%r, root=%r)>' % (
      self.__class__.__name__, self.name, self.root
    )

  def __str__(self):
    return self.name

  @property
  def versioned_name(self):
    """Project name, including version if present."""
    return '%s-%s' % (self.name, self.version) if self.version else self.name

  @property
  def files(self):
    """Returns a list of tuples of files included in the project archive.

    The first element of each tuple is the absolute local path to the file, the
    second the path of the file in the archive.

    .. note::

      This property should not be used to add files. Use :meth:`add_file`
      instead.

    """
    return [(e[1][0], e[0]) for e in self._files.items()]

  @property
  def jobs(self):
    """Returns a dictionary of all jobs in the project, keyed by name.

    .. note::

      This property should not be used to add jobs. Use :meth:`add_job`
      instead.

    """
    return _JobDict(self._jobs)

  def add_file(self, path, archive_path=None, overwrite=False):
    """Include a file in the project archive.

    :param path: Path to file. If no project `root` exists, only absolute paths
      are allowed. Otherwise, this path can also be relative to said `root`.
    :param archive_path: Path to file in archive (defaults to same as `path`).
    :param overwrite: Allow overwriting any previously existing file in this
      archive path.

    If the current project has its `root` parameter specified, this method will
    allow relative paths (and join those with the project's `root`), otherwise
    it will throw an error. Furthermore, when a project `root` exists, adding
    files above it without specifying an `archive_path` will raise an error.
    This is done to avoid having files in the archive with lower level
    destinations than the base root directory.

    """
    if not osp.isabs(path):
      if not self.root:
        raise AzkabanError(
          'Relative path not allowed without specifying a project root: %r.'
          % (path, )
        )
      path = osp.join(self.root, path)
    # disambiguate (symlinks, pardirs, etc.)
    path = osp.realpath(path)
    if archive_path:
      frozen = True
    else:
      frozen = False
      if self.root:
        if not path.startswith(self.root):
          raise AzkabanError(
            'Cannot add a file outside of the project root directory without\n'
            'specifying an archive path: %r' % (path, )
          )
        archive_path = osp.relpath(path, self.root)
      else:
        archive_path = path
    # leading separator meaningless inside archive (trimmed automatically)
    archive_path = archive_path.lstrip('/')
    if (
      archive_path in self._files and
      self._files[archive_path][0] != path and
      not overwrite
    ):
      raise AzkabanError('Inconsistent duplicate file: %r.' % (path, ))
    if not osp.exists(path):
      raise AzkabanError('File not found: %r.' % (path, ))
    self._files[archive_path] = (path, frozen)
    self._logger.info('Added file %r as %r.', path, archive_path)

  def add_job(self, name, job, **kwargs):
    """Include a job in the project.

    :param name: Name assigned to job (must be unique).
    :param job: :class:`~azkaban.job.Job` instance.
    :param kwargs: Keyword arguments that will be forwarded to the
      :meth:`~azkaban.job.Job.on_add` handler.

    This method triggers the :meth:`~azkaban.job.Job.on_add` method on the
    added job (passing the project and name as arguments, along with any
    `kwargs`). The handler will be called right after the job is added.

    """
    if not job is self._jobs.get(name, job):
      raise AzkabanError('Inconsistent duplicate job: %r.' % (name, ))
    job.on_add(self, name, **kwargs)
    self._jobs[name] = job
    self._logger.info('Added job %r.', name)

  def merge_into(self, project, overwrite=False, unregister=False):
    """Merge one project with another.

    :param project: Target :class:`Project` to merge into.
    :param overwrite: Overwrite any existing files.
    :param unregister: Unregister project after merging it.

    The current project remains unchanged while the target project gains all
    the current project's jobs and files.

    """
    self._logger.debug('Merging into %r.', project)
    for name, job in self._jobs.items():
      project.add_job(name, job, merging=self)
    for archive_path, (path, frozen) in self._files.items():
      if frozen:
        # propagate the archive path
        project.add_file(path, archive_path=archive_path, overwrite=overwrite)
      else:
        # autogenerate a new archive_path
        project.add_file(path, overwrite=overwrite)
    if unregister:
      self._registry.pop(self.name)

  def build(self, path, overwrite=False):
    """Create the project archive.

    :param path: Destination path.
    :param overwrite: Don't throw an error if a file already exists at `path`.

    """
    self._logger.debug('Building.')
    # not using a with statement for compatibility with older python versions
    if osp.exists(path) and not overwrite:
      raise AzkabanError('Path %r already exists.' % (path, ))
    if not (len(self._jobs) or len(self._files)):
      raise AzkabanError('Building empty project.')
    writer = ZipFile(path, 'w')
    try:
      if self.properties:
        with temppath() as fpath:
          write_properties(flatten(self.properties), fpath)
          writer.write(fpath, 'project.properties')
      for name, job in self._jobs.items():
        with temppath() as fpath:
          job.build(fpath)
          writer.write(fpath, '%s.job' % (name, ))
      for archive_path, (fpath, _) in self._files.items():
        writer.write(fpath, archive_path)
    finally:
      writer.close()
    self._logger.info('Built as %s.', path)

  @classmethod
  def load(cls, path, new=False):
    """Load Azkaban projects from script.

    :param path: Path to python module.
    :param new: If set to `True`, only projects loaded as a consequence of
      calling this method will be returned.

    Returns a dictionary of :class:`~azkaban.project.Project`'s keyed by
    project name. Only registered projects (i.e. instantiated with
    `register=True`) can be discovered via this method.

    """
    if not path:
      raise ImportError('Invalid project module path: %r', path)
    path = osp.abspath(path)
    _logger.debug('Attempting to load projects from: %r', path)
    head, tail = osp.split(path.rstrip(os.sep))
    sys.path.insert(0, head)
    # make sure we can load the module
    _registry = cls._registry
    cls._registry = {}
    # reset the registry to let us find out exactly how many projects are
    # loaded, even if there are name clashes
    try:
      __import__(osp.splitext(tail)[0])
      _logger.debug(
        'Found %s projects from loading %s: %s',
        len(cls._registry), path, ', '.join(cls._registry),
      )
      collisions = set(cls._registry) & set(_registry)
      if collisions:
        _logger.warn(
          '%s project name collisions detected by loading %s: %s',
          len(collisions), path, ', '.join(collisions)
        )
      registry = cls._registry.copy()
      # shallow copy of registry
    finally:
      for name, project in _registry.items():
        cls._registry.setdefault(name, project)
        # restore registry, keeping the latest definition of each project if a
        # newer project had the same name
    if new:
      _logger.info(
        '%s new projects were loaded from %s: %s',
        len(registry), path, ', '.join(registry)
      )
      return registry
    else:
      _logger.info(
        '%s total projects are now registered after loading %s: %s',
        len(cls._registry), path, ', '.join(cls._registry)
      )
      return cls._registry.copy()



#!/usr/bin/env python
# encoding: utf-8

"""Test CLI."""

from azkaban.__main__ import _parse_project, main
from azkaban.util import AzkabanError
from contextlib import contextmanager
from nose.tools import *
from shutil import rmtree
from tempfile import mkdtemp
import imp
import os
import os.path as osp
import sys


class TestParseProject(object):

  index = 0
  # python modules are hard to reload, we go around this by making all module
  # names unique

  @contextmanager
  def _temp_project(self, module_name=None, projects=None):
    path = mkdtemp()
    _cwd = os.getcwd()
    try:
      os.chdir(path)
      if module_name:
        module_path = osp.join(path, '%s_%s.py' % (module_name, self.index))
        self.index += 1
        with open(module_path, 'w') as writer:
          # yay for code generation, oh scala
          writer.write('from azkaban import Project\n')
          for index, (name, register) in enumerate(projects or []):
            writer.write(
              'pj_%s = Project(%r, register=%r)\n'
              % (index, name, register)
            )
      else:
        module_path = None
      yield module_name
    finally:
      os.chdir(_cwd)
      rmtree(path)

  # def test_single_default(self):
  #   with _project('jobs', [('bar', True)]):
  #     name, project = _parse_project(None)
  #     eq_(name, 'bar')
  #     ok_(project is not None)

  # @raises(AzkabanError)
  # def test_missing_module_default(self):
  #   with _project():
  #     name, project = _parse_project(None)

  # @raises(AzkabanError)
  # def test_missing_project_default(self):
  #   with _project('jobs'):
  #     name, project = _parse_project(None)

  # @raises(AzkabanError)
  # def test_multiple_names_default(self):
  #   with _project('jobs', [('bar', True), ('foo', True)]):
  #     name, project = _parse_project(None)

  # def test_multiple_names_with_name(self):
  #   with _project('jobs', [('bar', True), ('foo', True)]):
  #     name, project = _parse_project('bar')
  #     eq_(name, 'bar')
  #     ok_(project is not None)

  # @raises(AzkabanError)
  # def test_multiple_names_with_missing_name(self):
  #   with _project('jobs', [('bar', True), ('foo', True)]):
  #     name, project = _parse_project('faa', True)

  # def test_multiple_names_with_missing_name(self):
  #   with _project('jobs', [('bar', True), ('foo', True)]):
  #     name, project = _parse_project('faa', False)
  #     eq_(name, 'faa')
  #     ok_(project is None)

  # def test_normal(self):
  #   eq_(_parse_project('foo'), ('foo', None))
  #   eq_(_parse_project('foo.py'), ('foo.py', None))
  #   eq_(_parse_project('foo/'), ('foo/', None))

  # @raises(ImportError)
  # def test_missing_module(self):
  #   eq_(_parse_project('foo/:bar'), ('bar', None))

  # @raises(AzkabanError)
  # def test_require_project(self):
  #   _parse_project(':bar', require_project=True)


class TestMain(object):

  pass # TODO: add test for the CLI



#!/usr/bin/env python
# encoding: utf-8

"""Utility module."""

from contextlib import contextmanager
from functools import wraps
from itertools import chain
from logging.handlers import TimedRotatingFileHandler
from mimetypes import guess_type
from os import close, remove
from os.path import exists, expanduser
from requests.packages.urllib3.filepost import choose_boundary
from six import b, string_types
from six.moves.configparser import (NoOptionError, NoSectionError,
  ParsingError, RawConfigParser)
from tempfile import gettempdir, mkstemp
from traceback import print_exc
import logging as lg
import os.path as osp
import re
import sys
import warnings as wr


_logger = lg.getLogger(__name__)


class AzkabanError(Exception):

  """Base error class."""

  def __init__(self, message, *args):
    super(AzkabanError, self).__init__(message % args if args else message)


class Adapter(lg.LoggerAdapter):

  """Logger adapter that includes a prefix to all messages.

  :param prefix: Prefix string.
  :param logger: Logger instance where messages will be logged.
  :param extra: Dictionary of contextual information, passed to the formatter.

  """

  def __init__(self, prefix, logger, extra=None):
    lg.LoggerAdapter.__init__(self, logger, extra)
    # not using super since `LoggerAdapter` is an old-style class in python 2.6
    self.prefix = prefix

  def process(self, msg, kwargs):
    """Adds a prefix to each message.

    :param msg: Original message.
    :param kwargs: Keyword arguments that will be forwarded to the formatter.

    """
    return '%s :: %s' % (self.prefix, msg), kwargs


class Config(object):

  """Configuration class.

  :param path: path to configuration file. If no file exists at that location,
    the configuration parser will be empty.

  """

  def __init__(self, path=expanduser('~/.azkabanrc')):
    self.parser = RawConfigParser()
    self.path = path
    if exists(path):
      try:
        self.parser.read(self.path)
      except ParsingError:
        raise AzkabanError('Invalid configuration file %r.', path)

  def save(self):
    """Save configuration parser back to file."""
    with open(self.path, 'w') as writer:
      self.parser.write(writer)

  def get_option(self, command, name, default=None):
    """Get default option value for a command.

    :param command: Command the option should be looked up for.
    :param name: Name of the option.
    :param default: Default value to be returned if not found in the
      configuration file. If not provided, will raise
      :class:`~azkaban.util.AzkabanError`.

    """
    try:
      return self.parser.get(command, 'default.%s' % (name, ))
    except (NoOptionError, NoSectionError):
      if default:
        return default
      else:
        raise AzkabanError(
          'No default %(name)s found in %(path)r for %(command)s.\n'
          'You can specify one by adding a `default.%(name)s` option in the '
          '`%(command)s` section.'
          % {'command': command, 'name': name, 'path': self.path}
        )

  def get_file_handler(self, command):
    """Add and configure file handler.

    :param command: Command the options should be looked up for.

    The default path can be configured via the `default.log` option in the
    command's corresponding section.

    """
    handler_path = osp.join(gettempdir(), '%s.log' % (command, ))
    try:
      handler = TimedRotatingFileHandler(
        self.get_option(command, 'log', handler_path),
        when='midnight', # daily backups
        backupCount=1,
        encoding='utf-8',
      )
    except IOError:
      wr.warn('Unable to write to log file at %s.' % (handler_path, ))
    else:
      handler_format = '[%(levelname)s] %(asctime)s :: %(name)s :: %(message)s'
      handler.setFormatter(lg.Formatter(handler_format))
      return handler


class MultipartForm(object):

  """Form allowing streaming.

  :param files: List of filepaths. For more control, each file can also be
    represented as a dictionary with keys `'path'`, `'name'`, and `'type'`.
  :param params: Optional dictionary of parameters that will be included in the
    form.
  :param callback: Arguments `cur_bytes`, `tot_bytes`, `index`.
  :param chunksize: Size of each streamed file chunk.

  Usage:

  .. code:: python

    from requests import post

    form = MultipartForm(files=['README.rst'])
    post('http://your.url', headers=form.headers, data=form)

  """

  def __init__(self, files, params=None, callback=None, chunksize=4096):
    self._boundary = choose_boundary()
    self._params = params
    self._files = [
      {'path': o} if isinstance(o, string_types) else o
      for o in files
    ]
    self._callback = callback
    self._chunksize = chunksize
    # generate content type header
    self.headers = {
      'Content-Type': 'multipart/form-data; boundary=%s' % (self._boundary, )
    }
    # prepare files
    self._files = []
    for opts in files:
      if isinstance(opts, string_types):
        opts = {'path': opts}
      opts.setdefault('name', opts.get('name') or osp.basename(opts['path']))
      opts.setdefault('type', opts.get('type') or guess_type(opts['name'])[0])
      self._files.append(opts)

  def __iter__(self):
    def _generator(callback=self._callback):
      """Overall generator. Note the callback caching."""
      # set up counters used in the callback
      cur_bytes = 0
      tot_bytes = self.size
      # start the content body with the form parameters
      if self._params:
        params_content = b(''.join(
          '%s%s' % (self._get_section_header(name), content)
          for name, content in self._params.items()
        ))
      else:
        params_content = b''
      yield params_content
      # follow up with the files
      if len(self._files) == 1:
        # simple case, only one file (included as any other form param)
        file_opts = self._files[0]
        yield b(self._get_section_header(
          name='file',
          filename=file_opts['name'],
          content_type=file_opts['type'],
        ))
        for chunk in stream_file(file_opts['path'], self._chunksize):
          cur_bytes += len(chunk)
          yield chunk
          if callback:
            callback(cur_bytes, tot_bytes, 0)
      else:
        # we need to group all files in a single multipart/mixed section
        file_boundary = choose_boundary()
        yield b(self._get_section_header(
          name='files',
          content_type='multipart/mixed; boundary=%s' % (file_boundary, )
        ))
        for index, file_opts in enumerate(self._files):
          yield b(self._get_section_header(
            filename=file_opts['name'],
            content_type=file_opts['type'],
          ))
          for chunk in stream_file(file_opts['path'], self._chunksize):
            cur_bytes += len(chunk)
            yield chunk
            if callback:
              callback(cur_bytes, tot_bytes, index)
        yield b('\r\n--%s--' % (file_boundary, ))
      yield b('\r\n--%s--\r\n' % (self._boundary, ))
    return _generator()

  @property
  def size(self):
    """Total size of all the files to be streamed.

    Note that this doesn't include the bytes used for the header and
    parameters.

    """
    return sum(osp.getsize(d['path']) for d in self._files)

  def _get_section_header(self, name=None, content_disposition='form-data',
    filename=None, content_type=None, boundary=None):
    """Non streamed. Returns a string."""
    return (
      '\r\n'
      '--%(b)s\r\n'
      'Content-Disposition: %(d)s%(n)s%(f)s\r\n'
      '%(t)s\r\n'
      % {
        'b': boundary or self._boundary,
        'd': content_disposition,
        'n': '; name="%s"' % (name, ) if name else '',
        'f': '; filename="%s"' % (filename, ) if filename else '',
        't': 'Content-Type: %s\r\n' % (content_type, ) if content_type else ''
      }
    )


@contextmanager
def temppath():
  """Create a temporary filepath.

  Usage::

    with temppath() as path:
      # do stuff

  Any file corresponding to the path will be automatically deleted afterwards.

  """
  (desc, path) = mkstemp()
  close(desc)
  remove(path)
  try:
    yield path
  finally:
    if exists(path):
      remove(path)

def catch(*error_classes):
  """Returns a decorator that catches errors and prints messages to stderr.

  :param error_classes: Error classes.
  :param log: Filepath to log file.

  Also exits with status 1 if any errors are caught.

  """
  def decorator(func):
    """Decorator."""
    @wraps(func)
    def wrapper(*args, **kwargs):
      """Wrapper. Finally."""
      try:
        return func(*args, **kwargs)
      except error_classes as err:
        _logger.error(err)
        sys.stderr.write('%s\n' % (err, ))
        sys.exit(1)
      except Exception: # catch all
        _logger.exception('Unexpected exception.')
        print_exc()
        sys.exit(1)
    return wrapper
  return decorator

def flatten(dct, sep='.'):
  """Flatten a nested dictionary.

  :param dct: Dictionary to flatten.
  :param sep: Separator used when concatenating keys.

  """
  def _flatten(dct, prefix=''):
    """Inner recursive function."""
    items = []
    for key, value in dct.items():
      new_prefix = '%s%s%s' % (prefix, sep, key) if prefix else key
      if isinstance(value, dict):
        items.extend(_flatten(value, new_prefix).items())
      else:
        items.append((new_prefix, value))
    return dict(items)
  return _flatten(dct)

def human_readable(size):
  """Transform size from bytes to human readable format (kB, MB, ...).

  :param size: Size in bytes.

  """
  for suffix in ['bytes', 'kB', 'MB', 'GB', 'TB']:
    if size < 1024.0:
      return '%3.1f%s' % (size, suffix)
    size /= 1024.0

def write_properties(options, path=None, header=None):
  """Write options to properties file.

  :param options: Dictionary of options.
  :param path: Path to file. Any existing file will be overwritten. Writes to
    stdout if no path is specified.
  :param header: Optional comment to be included at the top of the file.

  """
  lines = ('%s=%s\n' % t for t in sorted(options.items()))
  if header:
    lines = chain(['# %s\n' % (header, )], lines)
  if path:
    with open(path, 'w') as writer:
      for line in lines:
        writer.write(line)
  else:
    for line in lines:
      sys.stdout.write(line)

def read_properties(*paths):
  """Read options from a properties file and return them as a dictionary.

  :param \*paths: Paths to properties file. In the case of multiple definitions
    of the same option, the latest takes precedence.

  Note that not all features of `.properties` files are guaranteed to be
  supported.

  """
  comment_p = re.compile(r'\s*(?:#|!)')
  continuation_whitespace_p = re.compile(r'\\\n\s*')
  separator_p = re.compile(r'(?<!\\)\s*(?::|=|\s)\s*')
  separator_replacement_p = re.compile(r'\\(:|=|\s)')
  opts = {}
  for path in paths:
    if not osp.exists(path):
      raise AzkabanError('No properties file found at %s.', path)
    try:
      with open(path) as reader:
        contents = continuation_whitespace_p.sub('', reader.read())
        lines = (
          tuple(s.strip() for s in separator_p.split(line, 1))
          for line in contents.split('\n')
          if line.strip() and not comment_p.match(line)
        )
        opts.update(dict(
          (
            separator_replacement_p.sub(lambda m: m.group(1), t[0]),
            t[1] if len(t) == 2 else ''
          )
          for t in lines
        ))
    except Exception:
      raise AzkabanError('Unsupported properties file: %r', path)
  return opts

def stream_file(path, chunksize):
  """Get iterator over a file's contents.

  :param path: Path to file.
  :param chunksize: Bytes per chunk.

  """
  with open(path, 'rb') as reader:
    while True:
      chunk = reader.read(chunksize)
      if chunk:
        yield chunk
      else:
        break



#!/usr/bin/env python
# encoding: utf-8

"""Test Azkaban project module."""

from azkaban.project import *
from azkaban.job import Job
from azkaban.util import AzkabanError, flatten, temppath
from nose.tools import eq_, ok_, raises, nottest
from nose.plugins.skip import SkipTest
from os import pardir
from os.path import basename, dirname, expanduser, relpath, abspath, join
from requests import ConnectionError, post
from six.moves.configparser import RawConfigParser
from time import sleep, time
from zipfile import ZipFile


# filepaths for testing
FILEPATHS = [__file__, abspath(join(dirname(__file__), pardir, 'README.rst'))]


class _TestProject(object):

  def setup(self):
    self.project = Project('foo')


class TestProjectAddFile(_TestProject):

  def test_add_file(self):
    self.project.add_file(__file__, 'bar')
    eq_(self.project._files, {'bar': (__file__, True)})

  def test_add_relative_file(self):
    project = Project('foo', root=__file__)
    project.add_file('test_job.py')
    path = join(dirname(__file__), 'test_job.py')
    eq_(project._files, {'test_job.py': (path, False)})

  def test_add_relative_file_with_archive_path(self):
    project = Project('foo', root=__file__)
    project.add_file('test_job.py', 'bar')
    eq_(
      project._files,
      {'bar': (join(dirname(__file__), 'test_job.py'), True)}
    )

  @raises(AzkabanError)
  def test_add_relative_file_outside_root(self):
    project = Project('foo', root=__file__)
    project.add_file(FILEPATHS[1])

  @raises(AzkabanError)
  def test_missing_file(self):
    self.project.add_file('bar')

  @raises(AzkabanError)
  def test_add_relative_file_without_root(self):
    self.project.add_file(relpath(__file__))

  def test_add_duplicate_file(self):
    self.project.add_file(__file__)
    self.project.add_file(__file__)
    eq_(self.project._files, {__file__.lstrip('/'): (__file__, False)})

  def test_add_duplicate_file_with_archive_path(self):
    self.project.add_file(FILEPATHS[0], 'foo')
    self.project.add_file(FILEPATHS[0], 'foo')
    eq_(self.project._files, {'foo': (__file__, True)})

  @raises(AzkabanError)
  def test_add_inconsistent_duplicate_file(self):
    self.project.add_file(FILEPATHS[0], 'foo')
    self.project.add_file(FILEPATHS[1], 'foo')

  def test_add_inconsistent_duplicate_file_with_overwrite(self):
    self.project.add_file(FILEPATHS[0], 'foo')
    self.project.add_file(FILEPATHS[1], 'foo', overwrite=True)
    path = abspath(FILEPATHS[1])
    eq_(self.project._files, {'foo': (path, True)})

  def test_files(self):
    self.project.add_file(FILEPATHS[0], 'foo')
    self.project.add_file(FILEPATHS[1])
    files = [(FILEPATHS[1], FILEPATHS[1].lstrip('/')), (FILEPATHS[0], 'foo')]
    eq_(sorted(self.project.files), files)


class TestProjectAddJob(_TestProject):

  def test_add_job(self):
    class OtherJob(Job):
      test = None
      def on_add(self, project, name, **kwargs):
        self.test = (project.name, name, kwargs)
    job = OtherJob()
    self.project.add_job('bar', job, baz=48)
    eq_(job.test, ('foo', 'bar', {'baz': 48}))
    ok_('bar' in self.project.jobs)

  def test_add_duplicate_consistent_job(self):
    job = Job()
    self.project.add_job('bar', job)
    self.project.add_job('bar', job)

  @raises(AzkabanError)
  def test_add_duplicate_inconsistent_job(self):
    self.project.add_job('bar', Job())
    self.project.add_job('bar', Job())

  def test_job(self):
    job = Job()
    self.project.add_job('bar', job)
    eq_(self.project.jobs['bar'], job)

  @raises(AzkabanError)
  def test_add_job_to_property(self):
    job = Job()
    self.project.jobs['bar'] = job

  @raises(AzkabanError)
  def test_missing_job(self):
    self.project.jobs['bar']


class TestProjectMerge(_TestProject):

  def test_merge_project(self):
    job_bar = Job()
    self.project.add_job('bar', job_bar)
    self.project.add_file(FILEPATHS[0], 'bar')
    project2 = Project('qux')
    job_baz = Job()
    project2.add_job('baz', job_baz)
    project2.add_file(FILEPATHS[1], 'baz')
    project2.merge_into(self.project)
    eq_(self.project.name, 'foo')
    eq_(self.project._jobs, {'bar': job_bar, 'baz': job_baz})
    eq_(
      self.project._files,
      {'bar': (FILEPATHS[0], True), 'baz': (FILEPATHS[1], True)}
    )

  def test_merge_project_on_add_kwargs(self):
    class OtherJob(Job):
      test = None
      def on_add(self, project, name, **kwargs):
        self.test = kwargs
    job = OtherJob()
    project2 = Project('qux')
    project2.add_job('bar', job)
    project2.merge_into(self.project)
    eq_(job.test, {'merging': project2})

  def test_merge_project_with_different_roots(self):
    project2 = Project('qux', root=__file__)
    project2.add_file(__file__)
    project2.add_file(__file__, 'bar')
    project1 = Project('oth', root=dirname(dirname(abspath(__file__))))
    project2.merge_into(project1)
    eq_(sorted(project1._files), ['bar', join('test', basename(__file__))])


class TestProjectBuild(_TestProject):

  @raises(AzkabanError)
  def test_build_empty(self):
    with temppath() as path:
      self.project.build(path)

  def test_build_single_job(self):
    job = Job({'a': 2})
    self.project.add_job('bar', job)
    with temppath() as path:
      self.project.build(path)
      reader =  ZipFile(path)
      try:
        ok_('bar.job' in reader.namelist())
        eq_(reader.read('bar.job').decode('utf-8'), 'a=2\n')
      finally:
        reader.close()

  def test_build_with_file(self):
    self.project.add_file(__file__.rstrip('c'), 'this.py')
    with temppath() as path:
      self.project.build(path)
      reader = ZipFile(path)
      try:
        ok_('this.py' in reader.namelist())
        eq_(
          reader.read('this.py').decode('utf-8').split('\n')[0],
          '#!/usr/bin/env python'
        )
      finally:
        reader.close()

  def test_build_multiple_jobs(self):
    self.project.add_job('foo', Job({'a': 2}))
    self.project.add_job('bar', Job({'b': 3}))
    self.project.add_file(__file__, 'this.py')
    with temppath() as path:
      self.project.build(path)
      reader = ZipFile(path)
      try:
        ok_('foo.job' in reader.namelist())
        ok_('bar.job' in reader.namelist())
        ok_('this.py' in reader.namelist())
        eq_(reader.read('foo.job').decode('utf-8'), 'a=2\n')
      finally:
        reader.close()


class TestProjectProperties(_TestProject):

  def test_no_properties_by_default(self):
    self.project.add_job('foo', Job({'a': 2}))
    with temppath() as path:
      self.project.build(path)
      reader = ZipFile(path)
      try:
        eq_(reader.namelist(), ['foo.job'])
      finally:
        reader.close()

  def test_properties_if_defined(self):
    self.project.add_job('foo', Job({'a': 2}))
    self.project.properties = {'bar': 123}
    with temppath() as path:
      self.project.build(path)
      reader = ZipFile(path)
      try:
        eq_(sorted(reader.namelist()), ['foo.job', 'project.properties'])
        eq_(
          reader.read('project.properties').decode('utf-8'),
          'bar=123\n'
        )
      finally:
        reader.close()

  def test_properties_flattened(self):
    self.project.add_job('foo', Job({'a': 2}))
    self.project.properties = {'param': {'foo': 1, 'bar': 'baz'}}
    with temppath() as path:
      self.project.build(path)
      reader = ZipFile(path)
      try:
        eq_(sorted(reader.namelist()), ['foo.job', 'project.properties'])
        eq_(
          reader.read('project.properties').decode('utf-8'),
          'param.bar=baz\nparam.foo=1\n'
        )
      finally:
        reader.close()



#!/usr/bin/env python
# encoding: utf-8

"""Test Azkaban job module."""

from azkaban.job import *
from azkaban.project import Project
from azkaban.util import AzkabanError, temppath
from nose.tools import eq_, ok_, raises, nottest


class TestJob(object):

  def test_generate_simple(self):
    job = Job({'a': 1, 'b': {'c': 2, 'd': 3}})
    with temppath() as path:
      job.build(path)
      with open(path) as reader:
        eq_(reader.read(), 'a=1\nb.c=2\nb.d=3\n')

  def test_generate_with_defaults(self):
    defaults = {'b': {'d': 4}, 'e': 5}
    job = Job(defaults, {'a': 1, 'b': {'c': 2, 'd': 3}})
    with temppath() as path:
      job.build(path)
      with open(path) as reader:
        eq_(reader.read(), 'a=1\nb.c=2\nb.d=3\ne=5\n')

  def test_generate_with_dependencies(self):
    foo = Job()
    bar = Job({'a': 3})
    job = Job({'a': 2, 'dependencies': 'bar,foo'})
    with temppath() as path:
      job.build(path)
      with open(path) as reader:
        eq_(reader.read(), 'a=2\ndependencies=bar,foo\n')

  def test_join_options(self):
    job = Job({'bar': range(3)})
    job.join_option('bar', ',')
    eq_(job.options['bar'], '0,1,2')

  def test_join_options_with_custom_formatter(self):
    job = Job({'bar': range(3)})
    job.join_option('bar', ' ', '(%s)')
    eq_(job.options['bar'], '(0) (1) (2)')

  def test_join_missing_option(self):
    job = Job({'bar': '1'})
    job.join_option('baz', ' ', '(%s)')
    eq_(job.options['bar'], '1')
    ok_(not 'baz' in job.options)

  def test_join_prefix(self):
    job = Job({'bar': {'a': 1, 'b.c': 'foo'}})
    job.join_prefix('bar', ',', '%s-%s')
    eq_(job.options['bar'], 'a-1,b.c-foo')



#!/usr/bin/env python
# encoding: utf-8

"""Test Azkaban util module."""

from azkaban.util import *
from contextlib import contextmanager
from nose.tools import eq_, ok_, raises, nottest
from six import u


class TestFlatten(object):

  def test_empty(self):
    eq_(flatten({}), {})

  def test_simple(self):
    dct = {'a': 1, 'B': 2}
    eq_(flatten(dct), dct)

  def test_nested(self):
    dct = {'a': 1, 'b': {'c': 3}}
    eq_(flatten(dct), {'a': 1, 'b.c': 3})


class TestConfig(object):

  @raises(AzkabanError)
  def test_invalid_file(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('a = 1\n')
      Config(path)

  def test_parser_get(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('[alias]\nfoo = 1\n')
      config = Config(path)
      eq_(config.parser.get('alias', 'foo'), '1')

  def test_save(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('[alias]\nfoo = 1\n')
      config = Config(path)
      config.parser.set('alias', 'bar', 'hello')
      config.save()
      same_config = Config(path)
      eq_(same_config.parser.get('alias', 'bar'), 'hello')

  def test_get_default_option_when_exists(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('[cmd]\ndefault.opt = foo\ndefault.bar = hi\n')
      config = Config(path)
      eq_(config.get_option('cmd', 'opt'), 'foo')

  @raises(AzkabanError)
  def test_get_default_option_when_option_is_missing(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('[cmd]\ndefault.opt = foo\ndefault.bar = hi\n')
      config = Config(path)
      config.get_option('cmd', 'opt2')

  @raises(AzkabanError)
  def test_get_default_option_when_section_is_missing(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('[cmd]\ndefault.opt = foo\ndefault.bar = hi\n')
      config = Config(path)
      config.get_option('cmd2', 'opt2')


class TestMultipartForm(object):

  def get_form_content(self, form):
    return b''.join(chunk for chunk in form)

  def test_single_file(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('HAI')
      form = MultipartForm([
        {'path': path, 'name': 'foo', 'type': 'text/plain'}
      ])
      ok_(b'HAI' in self.get_form_content(form))

  def test_multiple_files(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('HAI')
      form = MultipartForm([
        {'path': path, 'name': 'foo', 'type': 'text/plain'},
        path,
      ])
      ok_(b'HAI' in u(self.get_form_content(form)))

  def test_params(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('HAI')
      form = MultipartForm(
        files=[{'path': path, 'name': 'foo', 'type': 'text/plain'}, path],
        params={'foo': 'bar'},
      )
      content = self.get_form_content(form)
      ok_(b'name="foo"' in content)
      ok_(b'bar' in content)


class TestReadProperties(object):

  @staticmethod
  @contextmanager
  def temp_properties(contents):
    with temppath() as tpath:
      with open(tpath, 'w') as writer:
        writer.write(contents)
      yield tpath

  def test_separators(self):
    contents = 'a=1\nb=2\nc  4'
    with self.temp_properties(contents) as path:
      eq_(read_properties(path), {'a': '1', 'b': '2', 'c': '4'})

  def test_comments(self):
    contents = '# fee\na=1\nb=2\n   !faye\n'
    with self.temp_properties(contents) as path:
      eq_(read_properties(path), {'a': '1', 'b': '2'})

  def test_escaped_newline(self):
    contents = 'a = bo\\\n    hi\nb = 2\n'
    with self.temp_properties(contents) as path:
      eq_(read_properties(path), {'a': 'bohi', 'b': '2'})

  def test_empty_value(self):
    contents = 'a = b\nc\nd : 4'
    with self.temp_properties(contents) as path:
      eq_(read_properties(path), {'a': 'b', 'c': '', 'd': '4'})

  def test_escaped_separators(self):
    contents = 'a\=b = 5\nfoo\ bar :ja\n'
    with self.temp_properties(contents) as path:
      eq_(read_properties(path), {'a=b': '5', 'foo bar': 'ja'})



#!/usr/bin/env python
# encoding: utf-8

"""AzkabanPig: an extension for pig scripts to Azkaban CLI.

Usage:
  azkabanpig PATH ...
             [-j JAR] ... [-o OPTION] ...
             [-bp PROJECT] [-t TYPE] [-a ALIAS | -u URL]
  azkabanpig -h | --help | -l | --log

Arguments:
  PATH                          Path to pig script. If more than one path is
                                specified, they will be run in order. Note that
                                all the pig jobs will share the same Azkaban
                                options; for more flexibility, you can use
                                AzkabanCLI directly.

Options:
  -a ALIAS --alias=ALIAS        Cf. `azkaban --help`.
  -b --background               Run job asynchronously. AzkabanPig will launch
                                the workflow and return.
  -h --help                     Show this message and exit.
  -j JAR --jar=JAR              Path to jar file. It will be available on the
                                class path when the pig script is run, no need
                                to register it inside your scripts.
  -l --log                      Show path to current log file and exit.
  -o OPTION --option=OPTION     Azkaban option. Can either be a path to a
                                properties file or a single option formatted as
                                key=value. E.g. `-o param.foo=bar` will
                                substitute parameter `$foo` with `'bar'` in the
                                pig script.
  -p PROJECT --project=PROJECT  Project name under which to run the pig script.
  -t TYPE --type=TYPE           Pig job type used.
  -u URL --url=URL              Cf. `azkaban --help`.

Examples:
  azkabanpig my_script.pig
  azkabanpig -ba foo first_script.pig second_script.pig
  azkabanpig -u http://url.to.azkaban -o param.my_output=bar.dat foo.pig

AzkabanPig returns with exit code 1 if an error occurred and 0 otherwise.

"""

__all__ = ['PigJob']

from docopt import docopt
from os import sep
from os.path import abspath, basename, exists
from time import sleep
from ..job import Job
from ..project import Project
from ..remote import Execution, Session
from ..util import AzkabanError, Config, catch, temppath
import logging as lg
import sys


class PigJob(Job):

  """Convenience job class for running pig scripts.

  :param options: Tuple of options (cf. :class:`~azkaban.job.Job`). These
    options must specify a `'pig.script'` key. The corresponding file will
    then automatically be included in the project archive.

  This class allows you to specify JVM args as a dictionary by correctly
  converting these to the format used by Azkaban when building the job options.
  For example: `{'jvm.args': {'foo': 1, 'bar': 2}}` will be converted to
  `jvm.args=-Dfoo=1 -Dbar=2`. Note that this enables JVM args to behave like
  all other `Job` options when defined multiple times (latest values taking
  precedence).

  Finally, by default the job type will be set automatically to `'pig'`. You
  can also specify a custom job type for all :class:`PigJob` instances in the
  `azkabanpig` section of the `~/.azkabanrc` configuration file via the
  `default.type` option.

  """

  def __init__(self, *options):
    super(PigJob, self).__init__(
      {'type': Config().get_option('azkabanpig', 'type', 'pig')},
      *options
    )
    try:
      self.path = self.options['pig.script']
    except KeyError:
      raise AzkabanError('Missing `\'pig.script\'` option.')
    else:
      # absolute archive paths get trimmed
      self.options['pig.script'] = self.path.lstrip('/')
      self.join_prefix('jvm.args', ' ', '-D%s=%s')

  def on_add(self, project, name, **kwargs):
    """This handler adds the corresponding script file to the project."""
    if not 'merging' in kwargs:
      project.add_file(self.path)


class _PigProject(Project):

  """Project to run pig jobs from the command line.

  :param name: Project name used.
  :param paths: Paths to pig scripts, these will be run in order.
  :param pig_type: Pig job type used.

  """

  def __init__(self, name, paths, pig_type=None):
    super(_PigProject, self).__init__(name, register=False)
    self.ordered_jobs = [basename(path) for path in paths]
    for path, dep in zip(paths, [None] + self.ordered_jobs):
      options = {'pig.script': abspath(path)}
      if pig_type:
        options['type'] = pig_type
      options['dependencies'] = dep or '' # override key if exists
      self.add_job(basename(path), PigJob(options))

  def logs(self, execution, delay=10):
    """Jobs logs. In order.

    :param execution: `azkaban.remote.Execution`
    :param delay: Poll delay (in seconds) while job is preparing.

    """
    ok_statuses = set(['RUNNING', 'SUCCEEDED'])
    for job in self.ordered_jobs:
      while execution.status['status'] == 'PREPARING':
        # Delay log query until job is done preparing otherwise the log file
        # won't exist yet (and the server will send a 500 back).
        self._logger.warning('Job %s preparing.', job)
        sleep(delay)
      for line in execution.job_logs(job):
        yield line
      if not execution.status['status'] in ok_statuses:
        raise AzkabanError('Job %s failed.', job)
      else:
        self._logger.info('Job %s finished.', job)


@catch(AzkabanError)
def main():
  """AzkabanPig entry point."""
  args = docopt(__doc__)
  # activate logging
  logger = lg.getLogger()
  logger.setLevel(lg.DEBUG)
  handler = Config().get_file_handler('azkabanpig')
  if handler:
    logger.addHandler(handler)
  # capture pesky unverified requests warnings
  lg.captureWarnings(True)
  # handle this command separately
  if args['--log']:
    if handler:
      sys.stdout.write('%s\n' % (handler.baseFilename, ))
      sys.exit(0)
    else:
      raise AzkabanError('No log file active.')
  # create project
  paths = args['PATH']
  jars = args['--jar'] or []
  session = Session(args['--url'], args['--alias'])
  project = _PigProject(
    name=args['--project'] or Config().get_option('azkabanpig', 'project'),
    paths=paths,
    pig_type=args['--type'],
  )
  project.properties = {
    'user.to.proxy': session.user,
    'pig.additional.jars': ','.join(abspath(j).lstrip(sep) for j in jars),
  }
  project.properties.update(
    opt.split('=', 1)
    for opt in args['--option']
    if '=' in opt
  )
  job_paths = (opt for opt in args['--option'] if not '=' in opt)
  for i, job_path in enumerate(job_paths):
    if not exists(job_path):
      raise AzkabanError(
        'Invalid `--option`: %s\nOptions should point to an existing job file'
        ' or be of the form `key=value`.'
        % (job_path, )
      )
    project.add_file(abspath(job_path), '%s.properties' % (i, ))
  for jar in jars:
    project.add_file(abspath(jar))
  with temppath() as tpath:
    project.build(tpath)
    while True:
      try:
        session.upload_project(project, tpath)
      except AzkabanError:
        session.create_project(project, project)
      else:
        break
  res = session.run_workflow(project, basename(paths[-1]))
  exe = Execution(session, res['execid'])
  if args['--background']:
    sys.stdout.write('Execution running at %s\n' % (exe.url, ))
  else:
    try:
      for line in project.logs(exe):
        sys.stdout.write('%s\n' % (line.encode('utf-8'), ))
    except KeyboardInterrupt:
      choice = raw_input('\nCancel execution [yN]? ').lower()
      if choice and choice[0] == 'y':
        sys.stdout.write('Killing... ')
        exe.cancel()
        sys.stdout.write('Done.\n')
      else:
        sys.stdout.write('Execution still running at %s\n' % (exe.url, ))

if __name__ == '__main__':
  main()



#!/usr/bin/env python
# encoding: utf-8

"""Azkaban remote interaction module.

This contains the `Session` class which will be used for all interactions with
a remote Azkaban server.

"""

from .util import AzkabanError, Config, Adapter, MultipartForm, flatten
from getpass import getpass, getuser
from os.path import basename, exists
from requests.exceptions import HTTPError
from six import string_types
from six.moves.configparser import NoOptionError, NoSectionError
from six.moves.urllib.parse import urlparse
from time import sleep
import logging as lg
import requests as rq
import re


_logger = lg.getLogger(__name__)


def _azkaban_request(method, url, **kwargs):
  """Make request to azkaban server and catch common errors.

  :param method: GET, POST, etc.
  :param url: Endpoint url.
  :param **kwargs: Arguments forwarded to the request handler.

  This function is meant to handle common errors and return a more helpful
  message than the default one.

  """
  try:
    handler = getattr(rq, method.lower())
  except AttributeError:
    raise ValueError('Invalid HTTP method: %r.' % (method, ))
  else:
    try:
      response = handler(url, verify=False, **kwargs)
    except rq.ConnectionError:
      raise AzkabanError('Unable to connect to Azkaban server %r.', url)
    except rq.exceptions.MissingSchema:
      raise AzkabanError('Invalid Azkaban server url: %r.', url)
    else:
      return response

def _extract_json(response):
  """Extract JSON from Azkaban response, gracefully handling errors.

  :param response: Request response object.

  """
  try:
    json = response.json()
  except ValueError as err: # this should never happen
    _logger.error('No JSON decoded from response:\n%s', response.text)
    raise err
  else:
    if 'error' in json:
      raise AzkabanError(json['error'])
    elif json.get('status') == 'error':
      raise AzkabanError(json['message'])
    else:
      return json

def _parse_url(url):
  """Parse url, returning tuple of (username, password, address)

  :param url: HTTP endpoint (including protocol, port, and optional user /
    password).

  Supported url formats:

  + protocol://host:port
  + protocol://user@host:port
  + protocol://user:password@host:port
  + user@protocol://host:port (compatibility with older versions)
  + user:password@protocol://host:port (compatibility with older versions)

  """
  if not re.match(r'[a-zA-Z]+://', url) and not re.search(r'@[a-zA-Z]+://', url):
    # no scheme specified, default to http://
    url = 'http://' + url
  if re.search(r'@[a-zA-Z]+://', url):
    # compatibility mode: `user@protocol://host:port` or
    # `user:password@protocol://host:port`
    splitted = url.rstrip('/').split('@')
    if len(splitted) == 1:
      address = splitted[0]
      user = None
      password = None
    elif len(splitted) == 2:
      address = splitted[1]
      creds = splitted[0].split(':', 1)
      if len(creds) == 1:
        user = creds[0]
        password = None
      else:
        user, password = creds
    else:
      raise AzkabanError('Malformed url: %r' % (url, ))
    return user, password, address
  else:
    parsed = urlparse(url)
    return (parsed.username, parsed.password,
            '%s://%s:%s' % (parsed.scheme, parsed.hostname, parsed.port))

def _resolve_alias(config, alias):
  """Get url associated with an alias.

  :param alias: Alias name.

  """
  try:
    return config.parser.get('alias', alias)
  except (NoOptionError, NoSectionError):
    raise AzkabanError('Alias %r not found.', alias)

def _get_session_id(config, url):
  """Retrieve session id associated with url.

  :param url: HTTP endpoint (including protocol, port and optional user).

  """
  try:
    return config.parser.get('session_id', url)
  except NoSectionError:
    config.parser.add_section('session_id')
  except NoOptionError:
    pass


class Session(object):

  """Azkaban session.

  :param url: HTTP endpoint (including protocol, port and optional user).
  :param alias: Alias name.
  :param attempts: Maximum number of attempts to refresh session.

  This class contains mostly low-level methods that translate directly into
  Azkaban API calls. The :class:`~azkaban.remote.Execution` class should be
  preferred for interacting with workflow executions.

  Note that each session's ID is lazily updated. In particular, instantiating
  the :class:`Session` doesn't guarantee that its current ID (e.g. loaded from
  the configuration file) is valid.

  """

  def __init__(self, url=None, alias=None, attempts=3):
    self.attempts = attempts
    self.config = Config()
    if not url:
      alias = alias or self.config.get_option('azkaban', 'alias')
      url = _resolve_alias(self.config, alias)
    self.user, self.password, self.url = _parse_url(url)
    if not self.user:
      self.user = getuser()
    self.id = _get_session_id(self.config, str(self).replace(':', '.'))
    self._logger = Adapter(repr(self), _logger)
    self._logger.debug('Instantiated.')

  def __repr__(self):
    return '<%s(url=\'%s@%s\')>' % (
      self.__class__.__name__, self.user, self.url
    )

  def __str__(self):
    return '%s@%s' % (self.user, self.url)

  def is_valid(self, response=None):
    """Check if the current session ID is valid.

    :param response: If passed, this reponse will be used to determine the
      validity of the session. Otherwise a simple test request will be emitted.

    """
    self._logger.debug('Checking if current session is valid.')
    if not self.id:
      self._logger.debug('No previous ID found.')
      return False
    if response is None:
      # issue a request to check if the ID is valid (note the explicit `None`
      # check as 500 responses are falsish).
      self._logger.debug('Checking if ID %s is valid.', self.id)
      response = _azkaban_request(
        'POST',
        '%s/manager' % (self.url, ),
        data={'session.id': self.id},
      )
      # the above request will return a 200 empty response if the current
      # session ID is valid and a 500 response otherwise
    if (
      '<!-- /.login -->' in response.text or # usual non API error response
      'Login error' in response.text or # special case for API
      '"error" : "session"' in response.text # error when running a flow's jobs
    ):
      self._logger.debug('ID %s is invalid:\n%s', self.id, response.text)
      return False
    else:
      self._logger.debug('ID %s is valid.', self.id)
      return True

  def get_workflow_executions(self, project, flow, start=0, length=10):
    """Fetch executions of a flow.

    :param project: Project name.
    :param flow: Flow name.
    :param start: Start index (inclusive) of the returned list.
    :param length: Max length of the returned list.

    """
    self._logger.debug('Fetching executions of %s/%s.', project, flow)
    res = self._request(
      method='GET',
      endpoint='manager',
      params={
        'ajax': 'fetchFlowExecutions',
        'project': project,
        'flow': flow,
        'start': start,
        'length': length
      },
    )
    if not res.text:
      # Azkaban returns a 200 empty response if the project doesn't exist so
      # we throw an explicit error here, rather than letting `_extract_json`
      # fail generically.
      raise AzkabanError(
        'Unable to fetch executions. Check that project %r exists.', project
      )
    else:
      return _extract_json(res)

  def get_running_workflows(self, project, flow):
    """Get running executions of a flow.

    :param project: Project name.
    :param flow: Flow name.

    Note that if the project doesn't exist, the Azkaban server will return a
    somewhat cryptic error `Project 'null' not found.`, even though the name of
    the project isn't `null`.

    """
    self._logger.debug('Fetching running executions of %s/%s.', project, flow)
    return _extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'ajax': 'getRunning',
        'project': project,
        'flow': flow,
      },
    ))

  def get_execution_status(self, exec_id):
    """Get status of an execution.

    :param exec_id: Execution ID.

    """
    self._logger.debug('Fetching status for execution %s.', exec_id)
    return _extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'execid': exec_id,
        'ajax': 'fetchexecflow',
      },
    ))

  def get_execution_logs(self, exec_id, offset=0, limit=50000):
    """Get execution logs.

    :param exec_id: Execution ID.
    :param offset: Log offset.
    :param limit: Size of log to download.

    """
    self._logger.debug('Fetching logs for execution %s.', exec_id)
    return _extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'execid': exec_id,
        'ajax': 'fetchExecFlowLogs',
        'offset': offset,
        'length': limit,
      },
    ))

  def get_job_logs(self, exec_id, job, offset=0, limit=50000):
    """Get logs from a job execution.

    :param exec_id: Execution ID.
    :param job: Job name.
    :param offset: Log offset.
    :param limit: Size of log to download.

    """
    self._logger.debug('Fetching logs for execution %s, job %s.', exec_id, job)
    return _extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'execid': exec_id,
        'jobId': job,
        'ajax': 'fetchExecJobLogs',
        'offset': offset,
        'length': limit,
      },
    ))

  def cancel_execution(self, exec_id):
    """Cancel workflow execution.

    :param exec_id: Execution ID.

    """
    self._logger.debug('Cancelling execution %s.', exec_id)
    res = _extract_json(self._request(
      method='GET',
      endpoint='executor',
      params={
        'execid': exec_id,
        'ajax': 'cancelFlow',
      },
    ))
    if 'error' in res:
      raise AzkabanError('Execution %s is not running.', exec_id)
    else:
      self._logger.info('Execution %s cancelled.', exec_id)
    return res

  def create_project(self, name, description):
    """Create project.

    :param name: Project name.
    :param description: Project description.

    """
    self._logger.debug('Creating project %s.', name)
    return _extract_json(self._request(
      method='POST',
      endpoint='manager',
      data={
        'action': 'create',
        'name': name,
        'description': description,
      },
    ))

  def delete_project(self, name):
    """Delete a project on Azkaban.

    :param name: Project name.

    """
    self._logger.debug('Deleting project %s.', name)
    res = self._request(
      method='GET',
      endpoint='manager',
      params={
        'project': name,
        'delete': 'true',
      },
    )
    msg = "Project '%s' was successfully deleted" % (name, )
    if not msg in res.text:
      raise AzkabanError('Delete failed. Check permissions and existence.')
    return res

  def run_workflow(self, name, flow, jobs=None, concurrent=True,
    properties=None, on_failure='finish', notify_early=False, emails=None):
    """Launch a workflow.

    :param name: Name of the project.
    :param flow: Name of the workflow.
    :param jobs: List of names of jobs to run (run entire workflow by default).
    :param concurrent: Run workflow concurrently with any previous executions.
    :param properties: Dictionary that will override global properties in this
      execution of the workflow. This dictionary will be flattened similarly to
      how :class:`~azkaban.job.Job` options are handled.
    :param on_failure: Set the execution behavior on job failure. Available
      options: `'finish'` (finish currently running jobs, but do not start any
      others), `'continue'` (continue executing jobs as long as dependencies
      are met),`'cancel'` (cancel all jobs immediately).
    :param notify_early: Send any notification emails when the first job fails
      rather than when the entire workflow finishes.
    :param emails: List of emails or pair of list of emails to be notified
      when the flow fails. Note that this will override any properties set in
      the worfklow. If a single list is passed, the emails will be used for
      both success and failure events. If a pair of lists is passed, the first
      will receive failure emails, the second success emails.

    Note that in order to run a workflow on Azkaban, it must already have been
    uploaded and the corresponding user must have permissions to run it.

    """
    self._logger.debug('Starting project %s workflow %s.', name, flow)
    request_data = {
      'ajax': 'executeFlow',
      'project': name,
      'flow': flow
    }
    request_data.update(self._run_options(
      name,
      flow,
      jobs=jobs,
      concurrent=concurrent,
      properties=properties,
      on_failure=on_failure,
      notify_early=notify_early,
      emails=emails
    ))
    res = _extract_json(self._request(
      method='POST',
      endpoint='executor',
      include_session='params',
      data=request_data,
    ))
    self._logger.info('Started project %s workflow %s.', name, flow)
    return res

  def schedule_workflow(self, name, flow, date, time, period=None, **kwargs):
    """Schedule a workflow.

    :param name: Project name.
    :param flow: Name of flow in project.
    :param date: Date of the first run (possible values:
      `'08/07/2014'`, `'12/11/2015'`).
    :param time: Time of the schedule (possible values:
      `'9,21,PM,PDT'`, `'10,30,AM,PDT'`).
    :param period: Frequency to repeat. Consists of a number and a unit
      (possible values: `'1s'`, `'2m'`, `'3h'`, `'2M'`). If not specified
      the flow will be run only once.
    :param \*\*kwargs: See :meth:`run_workflow` for documentation.

    """
    self._logger.debug('Scheduling project %s workflow %s.', flow, name)
    request_data = {
      'ajax': 'scheduleFlow',
      'projectName': name,
      'projectId': self._get_project_id(name),
      'flow': flow,
      'scheduleDate': date,
      'scheduleTime': time,
      'is_recurring': 'on' if period else 'off',
    }
    if period:
      request_data['period'] = period
    request_data.update(self._run_options(name, flow, **kwargs))
    res = _extract_json(self._request(
      method='POST',
      endpoint='schedule',
      data=request_data,
    ))
    self._logger.info('Scheduled project %s workflow %s.', name, flow)
    return res

  def unschedule_workflow(self, name, flow):
    """Unschedule a workflow.

    :param name: Project name.
    :param flow: Name of flow in project.

    """
    self._logger.debug('Unscheduling project %s workflow %s.', flow, name)
    request_data = {
      'action': 'removeSched',
      'scheduleId': self.get_schedule(name, flow)['scheduleId'],
    }
    res = _extract_json(self._request(
      method='POST',
      endpoint='schedule',
      data=request_data,
    ))
    self._logger.info('Unscheduled project %s workflow %s.', name, flow)
    return res

  def get_schedule(self, name, flow):
    """Get schedule information.

    :param name: Project name.
    :param flow: Name of flow in project.

    """
    self._logger.debug(
      'Retrieving schedule for project %s workflow %s.', flow, name
    )
    res = _extract_json(self._request(
      method='GET',
      endpoint='schedule',
      params={
        'ajax': 'fetchSchedule',
        'projectId': self._get_project_id(name),
        'flowId': flow,
      },
    ))
    self._logger.info(
      'Retrieved schedule for project %s workflow %s.', name, flow
    )
    if 'schedule' not in res:
      raise AzkabanError(
        'Failed to get schedule. Check that the schedule exists.'
      )
    return res['schedule']

  def _get_project_id(self, name):
    """Fetch the id of a project.

    :param name: Project name.

    """
    self._logger.debug('Retrieving id for project %s.', name)
    try:
      res = _extract_json(self._request(
        method='GET',
        endpoint='manager',
        params={
          # there is no endpoint to get the project id, getPermissions is
          # the least expensive endpoint whose response contains the id
          'ajax': 'getPermissions',
          'project': name,
        },
      ))
    except ValueError:
      # Azkaban server sends a 200 empty response if the project doesn't exist
      raise AzkabanError(
        'Failed to get project id. Check that the project exists.'
      )
    else:
      project_id = res['projectId']
    self._logger.info('Retrieved id for project %s: %s.', name, project_id)
    return project_id

  def upload_project(self, name, path, archive_name=None, callback=None):
    """Upload project archive.

    :param name: Project name.
    :param path: Local path to zip archive.
    :param archive_name: Filename used for the archive uploaded to Azkaban.
      Defaults to `basename(path)`.
    :param callback: Callback forwarded to the streaming upload.

    """
    self._logger.debug('Uploading archive %r to project %s.', path, name)
    if not exists(path):
      raise AzkabanError('Unable to find archive at %r.' % (path, ))
    if not self.is_valid():
      self._refresh() # ensure that the ID is valid
    archive_name = archive_name or basename(path)
    form = MultipartForm(
      files=[{
        'path': path,
        'name': archive_name,
        'type': 'application/zip' # force this (tempfiles don't have extension)
      }],
      params={
        'ajax': 'upload',
        'project': name,
        'session.id': self.id,
      },
      callback=callback
    )
    # note that we have made sure the ID is valid, for two reasons:
    # + to avoid reuploading large files
    # + to simplify the custom ID update process (form parameter)
    res = _extract_json(self._request(
      method='POST',
      endpoint='manager',
      include_session=False,
      headers=form.headers,
      data=form,
    ))
    self._logger.info(
      'Archive %s for project %s uploaded as %s.', path, name, archive_name
    )
    return res

  def get_workflow_info(self, name, flow):
    """Get list of jobs corresponding to a workflow.

    :param name: Project name.
    :param flow: Name of flow in project.

    """
    self._logger.debug(
      'Fetching infos for workflow %s in project %s', flow, name
    )
    try:
      res = self._request(
        method='GET',
        endpoint='manager',
        params={
          'ajax': 'fetchflowjobs',
          'project': name,
          'flow': flow,
        },
      )
    except HTTPError:
      # the Azkaban server throws a NullPointerException if the flow doesn't
      # exist in the project, which causes a 500 response
      raise AzkabanError('Worklow %s not found in project %s.', flow, name)
    else:
      try:
        return _extract_json(res)
      except ValueError:
        # but sends a 200 empty response if the project doesn't exist
        raise AzkabanError('Project %s not found.', name)

  def _refresh(self, password=None):
    """Refresh session ID.

    :param password: Password used to log into Azkaban. If not specified,
      will prompt for one.

    Also caches the session ID for future use.

    """
    self._logger.debug('Refreshing.')
    attempts = self.attempts
    password = password or self.password
    while True:
      password = password or getpass('Azkaban password for %s: ' % (self, ))
      try:
        res = _extract_json(_azkaban_request(
          'POST',
          self.url,
          data={
            'action': 'login',
            'username': self.user,
            'password': password,
          },
        ))
      except AzkabanError as err:
        if not 'Incorrect Login.' in err.message:
          raise err
        self._logger.warning('Invalid login attempt.')
        attempts -= 1
        password = None
        if attempts <= 0:
          raise AzkabanError('Too many unsuccessful login attempts. Aborting.')
      else:
        break
    self.id = res['session.id']
    self.config.parser.set('session_id', str(self).replace(':', '.'), self.id)
    self.config.save()
    self._logger.info('Refreshed.')

  def _run_options(self, name, flow, jobs=None, concurrent=True,
    properties=None, on_failure='finish', notify_early=False, emails=None):
    """Construct data dict for run related actions.

    See :meth:`run_workflow` for parameter documentation.

    """
    if not jobs:
      disabled = '[]'
    else:
      all_names = set(
        n['id']
        for n in self.get_workflow_info(name, flow)['nodes']
      )
      run_names = set(jobs)
      missing_names = run_names - all_names
      if missing_names:
        raise AzkabanError(
          'Jobs not found in flow %r: %s.' %
          (flow, ', '.join(missing_names))
        )
      else:
        disabled = (
          '[%s]'
          % (','.join('"%s"' % (n, ) for n in all_names - run_names), )
        )
    try:
      failure_action = {
        'finish': 'finishCurrent',
        'continue': 'finishPossible',
        'cancel': 'cancelImmediately',
      }[on_failure]
    except KeyError:
      raise ValueError('Invalid `on_failure` value: %r.' % (on_failure, ))
    request_data = {
      'disabled': disabled,
      'concurrentOption': 'concurrent' if concurrent else 'skip',
      'failureAction': failure_action,
      'notifyFailureFirst': 'true' if notify_early else 'false',
    }
    if properties:
      request_data.update(dict(
        ('flowOverride[%s]' % (key, ), value)
        for key, value in flatten(properties).items()
      ))
    if emails:
      if isinstance(emails[0], string_types):
        failure_emails = ','.join(emails)
        success_emails = failure_emails
      else:
        failure_emails = ','.join(emails[0])
        success_emails = ','.join(emails[1])
      request_data.update({
        'failureEmails': failure_emails,
        'failureEmailsOverride': 'true',
        'successEmails': success_emails,
        'successEmailsOverride': 'true',
      })
    return request_data

  def _request(self, method, endpoint, include_session='cookies', **kwargs):
    """Make a request to Azkaban using this session.

    :param method: HTTP method.
    :param endpoint: Server endpoint (e.g. manager).
    :param include_session: Where to include the `session_id` (possible values:
      `'cookies'`, `'params'`, `False`).
    :param kwargs: Keyword arguments passed to :func:`_azkaban_request`.

    If the session expired, will prompt for a password to refresh.

    """
    full_url = '%s/%s' % (self.url, endpoint.lstrip('/'))

    if not self.id:
      self._logger.debug('No ID found.')
      self._refresh()

    def _send_request():
      """Try sending the request with the appropriate credentials."""
      if include_session == 'cookies':
        kwargs.setdefault('cookies', {})['azkaban.browser.session.id'] = self.id
      elif include_session == 'params':
        kwargs.setdefault('data', {})['session.id'] = self.id
      elif include_session:
        raise ValueError('Invalid `include_session`: %r' % (include_session, ))
      return _azkaban_request(method, full_url, **kwargs)

    response = _send_request()
    if not self.is_valid(response):
      self._refresh()
      response = _send_request()

    # `_refresh` raises an exception rather than letting an unauthorized second
    # request happen. this means that something is wrong with the server.
    if not self.is_valid(response):
      raise AzkabanError('Azkaban server is unavailable.')

    try:
      response.raise_for_status() # check that we get a 2XX response back
    except HTTPError as err: # catch, log, and reraise
      self._logger.warning(
        'Received invalid response from %s:\n%s',
        response.request.url, response.content
      )
      raise err
    else:
      return response


class Execution(object):

  """Remote workflow execution.

  :param session: :class:`Session` instance.
  :param exec_id: Execution ID.

  """

  def __init__(self, session, exec_id):
    self._session = session
    self.exec_id = exec_id

  @property
  def status(self):
    """Execution status."""
    return self._session.get_execution_status(self.exec_id)

  @property
  def url(self):
    """Execution URL."""
    return '%s/executor?execid=%s' % (self._session.url, self.exec_id)

  def cancel(self):
    """Cancel execution."""
    self._session.cancel_execution(self.exec_id)

  def logs(self, delay=5):
    """Execution log generator.

    :param delay: time in seconds between each server poll

    Yields line by line.

    """
    finishing = False
    offset = 0
    while True:
      logs = self._session.get_execution_logs(
        exec_id=self.exec_id,
        offset=offset,
      )
      if logs['length']:
        offset += logs['length']
        lines = (e for e in logs['data'].split('\n') if e)
        for line in lines:
          yield line
      elif finishing:
        break
      else:
        if self.status['status'] != 'RUNNING':
          finishing = True
      sleep(delay)

  def job_logs(self, job, delay=5):
    """Job log generator.

    :param job: job name
    :param delay: time in seconds between each server poll

    Yields line by line.

    """
    finishing = False
    offset = 0
    while True:
      try:
        logs = self._session.get_job_logs(
          exec_id=self.exec_id,
          job=job,
          offset=offset,
        )
      except HTTPError as err:
        # if Azkaban is hanging, the job might be stuck in preparing stage
        preparing = False
        while True:
          sleep(delay)
          preparing_jobs = set(
            e['id']
            for e in self.status['nodes']
            if e['status'] == 'PREPARING'
          )
          if job in preparing_jobs:
            if not preparing:
              preparing = True
              _logger.debug(
                'Job %s in execution %s is still preparing.', job, self.exec_id
              )
          else:
            break
        if not preparing:
          # something else is causing the error
          raise err
      else:
        if logs['length']:
          offset += logs['length']
          lines = (e for e in logs['data'].split('\n') if e)
          for line in lines:
            yield line
        elif finishing:
          break
        else:
          running_jobs = set(
            e['id']
            for e in self.status['nodes']
            if e['status'] == 'RUNNING'
          )
          if job not in running_jobs:
            finishing = True
      sleep(delay)

  @classmethod
  def start(cls, session, *args, **kwargs):
    """Convenience method to start a new execution.

    :param session: :class:`Session` instance.
    :param args: Cf. :meth:`Session.run_workflow`.
    :param kwargs: Cf. :meth:`Session.run_workflow`.

    """
    res = session.run_workflow(*args, **kwargs)
    return cls(session, res['execid'])



#!/usr/bin/env python

"""AzkabanCLI: a lightweight command line interface for Azkaban."""

from azkaban import __version__
from setuptools import find_packages, setup


setup(
    name='azkaban',
    version=__version__,
    description=__doc__,
    long_description=open('README.rst').read(),
    author='Matthieu Monsch',
    author_email='monsch@alum.mit.edu',
    url='http://azkabancli.readthedocs.org/',
    license='MIT',
    packages=find_packages(),
    classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python',
      'Programming Language :: Python :: 2.6',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.3',
    ],
    install_requires=[
      'six',
      'docopt',
      'requests>=2.0.1',
    ],
    entry_points={'console_scripts': [
      'azkaban = azkaban.__main__:main',
      'azkabanpig = azkaban.ext.pig:main',
    ]},
)
